-- ************************************************************
-- Fundamental Definition
-- ************************************************************

SUCCESS = 0;
WARNING = 1;
FALSE	= 1;
ERROR   = -1;
FATAL	= -1;
ABORT	= 2;

LO		= 16;
HI		= 32;
EQ		= 64;

SW_SHOWNORMAL = 1;

MSDE_INSTANCE_NAME 	= 'RSESM'
EXECUTE_WAITINGTIME = 180 * 1000
EXECUTE_WAITALWAYS	= -1

-- debug
function DebugOutput(level, group, info)
--[[
	logfilepath = string.format("%s\\Application Data\\Rising\\ESM\\setupscript.log", os.getenv("allusersprofile"));
	if (not FileSys.PathFileExists(swapfilepath)) and FileSys.CreatePath(string.format("%s\\Application Data\\Rising\\ESM", os.getenv("allusersprofile")))
	then
		local f = io.open(swapfilepath, "w");
		if f ~= nil then f:close() end
	end

	local f = io.open(logfilepath, "a")
	if f ~= nil then
		f:write( tostring(info).."\n" )
		f:close()
	end
]]--
	FileSys.RsLog(level, group, info)
end
-- debug


function creatqueue()
	return {top = 0, bottom = -1,
			push =  function (queue, data)
						queue[queue.top] = data;
						queue.top = queue.top + 1;
					end,
			pop  =  function (queue)
						if queue.top <= queue.bottom then return nil end
						queue.top = queue.top - 1;
						return queue[queue.top];
					end
			};
end


function _tostring(intend, value)
	if type(value) ~= "table" then
		return string.format("\"%s\"", tostring(value));
	end

	local result = "";
	for key, val in pairs(value) do
		if type(val) == "string" then
			val = string.gsub(val, "\\", "\\\\");
		end
		if type(key) == "number" then
			result = string.format("%s\n%s[%d] = %s,", result, intend, key, _tostring("\t"..intend, val));
		else
			result = string.format("%s\n%s%s = %s,", result, intend, key, _tostring("\t"..intend, val));
		end
	end
	return string.format("{\n%s%s\n%s}\n", intend, result, intend);
end


function _tablecopy (dst, src)
	if type(src) ~= "table" or type(dst) ~= "table" then
		return;
	end

	for key, value in pairs(src) do
		if type(value) == "table" then
			dst[key] = {}
			_tablecopy(dst[key], value)
		else
			dst[key] = value
		end
	end
end


function _createswapfile ()
	swapfilepath = string.format("%s\\Application Data\\Rising\\ESM\\swapfile", os.getenv("allusersprofile"));
--	local f = io.open(swapfilepath, "r");
	if not FileSys.PathFileExists(swapfilepath) then
		if TrackOp(FileSys.CreatePath(string.format("%s\\Application Data\\Rising\\ESM", os.getenv("allusersprofile"))), VerifyResult,
				   string.format("create swapfile[%s]", swapfilepath))
		then
			local f = io.open(swapfilepath, "w");
			if f ~= nil then f:close() end
		end
	end
end


function _loadswapfile ()
	local f = io.open(swapfilepath, "r");
	TrackOp(f ~= nil, VerifyResult, string.format("open swapfile[%s]", swapfilepath))
	local s = (f ~= nil and tostring(f:read("*a"))) or "";

	local status = pcall(function () g_tempdata = loadstring(string.format("return %s;", s))() or {} end);
	if not TrackOp(	status, VerifyResult, string.format("load swafile[%s: length=%d]", swapfilepath, string.len(s))
				   )
	then
		g_tempdata = {};		-- clean g_tempdata if any error occured in loading swapfile;
	end
	if f ~= nil then f:close() end;
end

function _updateswapfile ()
	local f = io.open(swapfilepath, "w");
	if not TrackOp(f ~= nil, VerifyResult, string.format("update swapfile[%s]", swapfilepath)) then
		return false;
	end
	f:write(_tostring("", g_tempdata));
	f:close();
	return true
end

-------------

function Initialize( clean )
	lasterror  = {};
	success	   = 0
	failed     = 0

	tracker	   = {};
	paramsourcelist = {[1] = "", [2] = "tempcfg_", [3] = "quick_", [4] = "local_"};
	component_configfilesuffixies = {"eptray", "rscom", "epsrv", "drvapp"}
	ipv4_pattern = "(%d+%.%d+%.%d+%.%d+:?%d*)"
	ipv6_pattern = "(%x+:%x+:%x+:%x+:%x+:%x+:%x+:%x+:?%d*)"

	undolist   = creatqueue();
	redolist   = creatqueue();

	donelastop = true;
	donerestart= false;
	noalert	   = false;
	doneabort  = false;

	-- make sure whether it is in debugmode
	local cmd 		= winfunc.GetCmdLine();
	local _, _, s	= string.find(cmd.lower(cmd), "/debug");
	debugmode		= (s == "/debug")

	Service	   = winfunc.WinService();
	Threading  = winfunc.WinThreading();
	MD5        = winfunc.WinCheckSumMD5();
	FileSys    = winfunc.WinFileSystems();
	RegKey     = winfunc.WinRegKey();
	IPChecker  = winfunc.IPValidater();
	WinDir 	   = winfunc.WinDir();
	WinIni 		= winfunc.WinIni();
	FileFind 	= winfunc.WinFileFind();
end


function SetLastError(err)
	table.insert(lasterror, tostring(err));
	failed = failed + 1;
end


function GetLastError()
	return table.concat(lasterror, "\n");
end


function VerifyResult(s)
	if 		type(s) == "number"  then
		return s > 0;
	elseif type(s) == "boolean" then
		return s;
	elseif type(s) == "nil" 	 then
		return false;
	elseif type(s) == "string"  then
		if tonumber(s) ~= nil then
			return tonumber > 0;
		else
			return s == "true";
		end
	end
end


function TrackOp (res, verify, err)
	if verify(res) == false then
		DebugOutput(0, 'basement', string.format("FAILED TO: (%s)", err))
		SetLastError(err);
		return false;
	end
	DebugOutput(1, 'basement', string.format("SUCCEED TO: (%s)", err))
	success = success + 1;
	return true;
end


function RecordOp (prefix, err)
	DebugOutput (1, 'basement', string.format("%s : (%s)", prefix, err))
end

function RecordErrOp (prefix, err)
	DebugOutput (1, 'basement', string.format("%s : (%s)", prefix, err))
	SetLastError(err);
end


function TrackReset()
	lasterror = {}
	failed    = 0
	success	  = 0
end


-- ************************************************************
-- Basic Atomic Operation, which should be undo if necessary
-- ************************************************************

function ATM_UpdateFileContent(src, content)
	local f = io.open(src, "w");
	if TrackOp(f ~= nil, VerifyResult, string.format("create file[%s] when updating its content", src)) then
		f:write(tostring(content));
		f:close();
		return true;
	end
	donelastop = false;
	return false;
end


function ATM_CopyFile(src, dest, attr)
	local err;
	if attr == nil then attr = 0x80 end   -- FILE_ATTRIBUTE_NORMAL in default
	for i = 1, 10, 1 do
		err = FileSys.CopyFile(src, dest, false, true, attr);	-- not failed if file exist, create folder
		if VerifyResult(err) then break
		else				 Threading.Sleep(1000) end
	end

	-- rename then copy if previous copy failed
	if not VerifyResult(err) then
		if FileSys.PathFileExists(dest) then
			local i 	= 0;
			local temp 	= string.format("%s.%s", dest, i);

			-- find valid temp file
			while FileSys.PathFileExists(temp) do
				err = FileSys.DeletePath(temp, true, false, true);
				if TrackOp(err, VerifyResult, string.format("try copy through tempfile[%s]", temp)) then
					break;
				end
				FileSys.MovePathEx(temp, true, false, true);		-- remove tempfile if occupied
				i 	 = i + 1;
				temp = string.format("%s.%s", dest, i);
			end

			-- try to copy through temp file
			err = FileSys.MoveFile(dest, temp)
				  and FileSys.CopyFile(src, dest, false, true, attr);
			if not TrackOp(err, VerifyResult, string.format("rename and replace file from [%s] to [%s]", src, dest)) then
				err = FileSys.MoveFileEx(src, dest, 0x04);	-- MOVEFILE_DELAY_UNTIL_REBOOT
				donerestart = err;
			end
		else
			err = FileSys.MoveFileEx(src, dest, 0x04);	-- MOVEFILE_DELAY_UNTIL_REBOOT
			donerestart = err;
		end
	end

	if not TrackOp(err, VerifyResult, string.format("copy file from [%s] to dest[%s]", src, dest)) then
		donelastop = false;
		return false;
	end
	undolist:push({ATM_DelPath, string.format("\"%s\"", dest)});
	return true;
end


function ATM_DelPath (dest, failaserror)
	if failaserror == nil then failaserror = true end

    if FileSys.PathFileExists(dest) then
		local err;
		for i = 1, 10, 1 do
			err = FileSys.DeletePath(dest, true, false, true);
			if VerifyResult(err) then break
			else				 Threading.Sleep(1000) end
		end

		if not VerifyResult(err) then
			donerestart = FileSys.MovePathEx(dest, true, false, true);

			if not TrackOp(donerestart, VerifyResult, string.format("delete directory[%s]", dest)) then
				if failaserror then
					donelastop = false;
					return false;
				end
			end
	    end
	end
	-- no undoes
	return true;
end


function ATM_CreatePath (dest)
	if (not FileSys.PathFileExists(dest)) and
		not TrackOp(FileSys.CreatePath(dest), VerifyResult, string.format("create path[%s]", dest))
	then
		donelastop = false;
		return false;
	end
	undolist:push({ATM_DelPath, dest});
	return true
end


function ATM_LoadTempData(name)
	_createswapfile();
	if g_tempdata == nil then _loadswapfile () end
	data = g_tempdata[name];
	RecordOp("SUCCEED TO", string.format("load temp data[%s : %s]", name, tostring(data)));
	return data;
end

function ATM_LoadTempDataFromSourceList(name)
	local r = nil;
	for i = 1, 4, 1 do
		r = ATM_LoadTempData(string.format("%s%s", paramsourcelist[i], name));
		if r ~= nil then break end
	end
	return r;
end

function ATM_SaveTempData(name, value)
	_createswapfile();
	_loadswapfile ()
	g_tempdata[name] = value;
	TrackOp(_updateswapfile(), VerifyResult, string.format("save temp data[%s : %s]", name, tostring(value)));
end

function ATM_DelTempData(name)
	_createswapfile();
	_loadswapfile ()
	if g_tempdata[name] ~= nil then
		g_tempdata[name] = nil;
	end
	TrackOp(_updateswapfile(), VerifyResult, string.format("delete temp data[%s]", name));
end


function ATM_CreateRegKey(hkey, subkey, name)
	local result  = true;
    if not TrackOp(RegKey.Create(hkey, string.format("%s\\%s", subkey, name), "", 0, "KEY_ALL_ACCESS"), VerifyResult,
					string.format("open regkey[%s//%s]", hkey, subkey)
				   )
	then
        donelastop = false;
		result	   = false;
    end
    RegKey.Close();
    undolist:push({ATM_RemoveRegKey, string.format("\"%s\"", hkey), string.format("\"%s\"", subkey), string.format("\"%s\"", name)});
    return result;
end


function ATM_SetRegKey(hkey, subkey, name, value, type)
	local result  = true;
    if not TrackOp(RegKey.Open(hkey, subkey, "KEY_WRITE"), VerifyResult, string.format("open regkey[%s//%s]", hkey, subkey))
       or not TrackOp(RegKey.SetStringValue(name, value, type), VerifyResult, string.format("set value[%s:%s] of regkey[%s//%s]", name, value, hkey, subkey))
    then
        donelastop = false;
		result	   = false;
    end
    RegKey.Close();
    undolist:push({ATM_RemoveRegKey, string.format("\"%s\"", hkey), string.format("\"%s\"", subkey), string.format("\"%s\"", name)});
    return result;
end


function ATM_GetRegKey(hkey, subkey, name)
	local result  = nil;
    if TrackOp(RegKey.Open(hkey, subkey, "KEY_READ"), VerifyResult, string.format("open regkey[%s//%s]", hkey, subkey)) then
		result = tostring(RegKey.QueryStringValue(name));
		if not TrackOp(result ~= "", VerifyResult, string.format("get value[%s] of regkey[%s//%s]", name, hkey, subkey)) then
			donelastop = false;
			result	   = nil;
		end
	end
    RegKey.Close();
    return result;
end

-- delete value
function ATM_RemoveRegKey(hkey, subkey, name)
	local result  = true;
    if not TrackOp(RegKey.Open(hkey, subkey, "KEY_READ_WRITE"), VerifyResult, string.format("open regkey[%s//%s]", hkey, subkey))
       or not TrackOp(RegKey.DeleteValue(name), VerifyResult, string.format("delete value[%s] of regkey[%s//%s]", name, hkey, subkey))
    then
        donelastop = false;
		result	   = false;
    end
    RegKey.Close();
    -- no undoinfo
    return result;
end

-- delete key
function ATM_RemoveRegKeyEx(hKey, subkey, keyname)
	local result  = true;
    if not TrackOp(RegKey.Open(hKey, subkey, "KEY_READ_WRITE"), VerifyResult, string.format("open regkey[%s\\%s]", hKey, subkey))
       or not TrackOp(RegKey.RecurseDeleteKey(keyname), VerifyResult, string.format("recurse delete key[%s] of regkey[%s\\%s]", keyname, hKey, subkey))
    then
        donelastop = false;
		result	   = false;
    end
    RegKey.Close();
    -- no undoinfo
    return result;
end

function ATM_CleanMoveFileExTracks (path)
	local hkey		= "HKEY_LOCAL_MACHINE"
	local subkey 	= "SYSTEM\\CurrentControlSet\\Control\\Session Manager";
	local value		= 'PendingFileRenameOperations';
	local result	= true;
	if not TrackOp(RegKey.Open(hkey, subkey, "KEY_READ_WRITE"), VerifyResult, string.format("open regkey[%s\\%s]", hkey, subkey)) then
		donelastop 	= false;
		result		= false;
	else
		if not TrackOp(RegKey.ModifyMultiStringValue(value, string.format("\\??\\%s", path)), VerifyResult,
						string.format("modify multistring[%s\\%s\\%s : %s]", hkey, subkey, value, path)
					   )
		then
			donelastop 	= false;
			result		= false;
		end
	end
	RegKey.Close();
	return result;
end


function ATM_UnzipFile (package, dstdir, pwd)
	local unzip  = winfunc.WinZip();
	--winfunc.msgbox(string.format("unzip file[%s] to dir[%s]", package, dstdir))
	if not TrackOp(unzip.UnZip(package, dstdir, pwd), VerifyResult, string.format("unzip file[%s] to dir[%s]", package, dstdir)) then
		local item 		= unzip.GetFailedItem();
		local temppath 	= string.format("%s\\esm.unzip.failed.retry.temp", dstdir)

		while item ~= -1 do
			local itempath 	= unzip.UnZipSpecItem(package, temppath, pwd, item);
			local err	  	= false;

			if TrackOp(itempath ~= nil, VerifyResult, string.format("unzip item[%dth] of file[%s]", item, package)) then
				itempath = string.format("%s\\%s", dstdir, itempath);

				local i 		 	= 0;
				local renamepath 	= string.format("%s.%s", itempath, i);

				-- find valid temp file
				while FileSys.PathFileExists(renamepath) do
					err = FileSys.DeletePath(renamepath, true, false, true);
					if TrackOp(err, VerifyResult, string.format("try delete discarded renamepath[%s]", renamepath)) then
						break;
					end
					FileSys.MovePathEx(renamepath, true, false, true);		-- remove tempfile if occupied
					i 	 	   = i + 1;
					renamepath = string.format("%s.%s", itempath, i);
				end

				if FileSys.PathFileExists(itempath) then
					err =  FileSys.MoveFile(itempath, renamepath);
					TrackOp(err, VerifyResult, string.format("rename file from [%s] to [%s]", itempath, renamepath))
				end

				err = FileSys.MoveFile(temppath, itempath);
				if not TrackOp(err, VerifyResult, string.format("replace file from [%s] to [%s]", temppath, itempath)) then
					if FileSys.PathFileExists(itempath) then
						RecordOp("SUCCEED TO", string.format("rename-replace failed, still use old file[%s]", itempath));
					else
						donelastop = false;		-- the case will be regarded as a failure, in which, unzip succeed, copy failed, and target file not exist
						return false;
					end
				else
					FileSys.MovePathEx(renamepath, true, false, true);		-- remove tempfile if occupied
				end
			else
				donelastop = false;		-- failed unzip will be regarded as a failure
				return false;
			end
			item = unzip.GetFailedItem();
		end
	end
	-- no undo info, since we don't know the reverse right exactly
	return true;
end


function ATM_ModifyPathSecurity (path, right, account, domain)
	local sid 		= winfunc.WinSid();
	local result   	= true;
	local fullname 	= (domain ~= nil and string.format("%s\\%s", domain, account)) or account;
	if not (TrackOp(sid.LoadAccount(fullname), VerifyResult,
					 string.format("load account[%s]", fullname)
					)
			 and TrackOp(winfunc.AddFileAccessRights(path, sid, right), VerifyResult,
						string.format("add right[%s] of account[%s] onto path[%s]", right, fullname, path)
					   )
			)
	then
		donelastop = false;
		result 	   = false;
	end
	-- no undo info, since we don't know the reverse right exactly
	return result;
end


function ATM_Execute(appfile, commandline, succeedcodes, waitforexit, waittime)
	if not waitforexit then succeedcodes = {0} end
	if waittime == nil then waittime = EXECUTE_WAITINGTIME end

	local showmode = (ATM_LoadTempData("showsubprocess_mode") == "true") or false;
	local exitcode = Threading.Execute(appfile, commandline, waitforexit, showmode, waittime);
	local err = false;

	if #succeedcodes > 0 then
		for i, k in pairs(succeedcodes) do
			if exitcode == k then err = true end
		end
	else
		err = true
	end
	if not TrackOp(err, VerifyResult, string.format("execute[%s %s] with exitcode[%s : %s]", appfile,
																							  commandline,
																							  tostring(exitcode),
																							  tostring(Threading.GetLastError())
													)
				   )
	then
		donelastop = false;
		return false;
	end
	return true;
end

function ATM_ExitProcessDuration (filepath, duration)
	local step   = 3;
	local remain = duration
	local exited = Threading.ProcessIsExit(filepath);
	while not exited do
		if remain < 0 then
			TrackOp(Threading.KillProcess(filepath), VerifyResult,
					string.format("kill process[%s] after waiting for [%d] secs", filepath, duration)
				   );
			break;
		end
		Threading.Sleep(step * 1000);
		remain 	 = remain - step;
		exited 	 = Threading.ProcessIsExit(filepath);
	end
	return true;
end

-- ************************************************************
-- Basic non-Atomic Operations, which need not undo
-- ************************************************************

function SelectFirstLocalIP ()
	local localips = winfunc.GetValidIPs();
	if TrackOp(type(localips) == 'string' and localips ~= "", VerifyResult, "fetch any valid host IPs(v4)") then
		local _, _, ip = string.find(localips, "^(%d+%.%d+%.%d+%.%d+),");
		if ip ~= nil then return ip end
	end
	return "localhost";
end

function CheckSPLMode (splfile)
	local result = "";
    repeat
        local doc = rapidxml.open (tostring(splfile));
	    if not TrackOp(doc ~= nil, VerifyResult, string.format("open splfile[%s]", splfile)) then	break end

	    local root = doc:root();
	    if root == nil or root:name() ~= 'subproductlist' then break end

	    result = root:attribute('mode') or "";
	until true
	return result;
end

function LoadCategory(category)
    local result = {};
    repeat
        local doc = rapidxml.open (tostring(category));
	    if not TrackOp(doc ~= nil, VerifyResult, string.format("open category[%s]", category)) then	break end

	    local root = doc:root();
	    if root == nil or root:name() ~= 'filelist' then break end

	    local files = root:select('file');
	    if files == nil or #files == 0 then break end

	    for inode, knode in pairs(files) do
	        local cel = {}
	        for iattr, kattr in pairs({'path', 'name', 'version', 'checksum', 'updatetime', 'rawsize', 'datafile'}) do
	            cel[kattr] = knode:attribute(kattr);
	        end
	        table.insert(result, cel);
	    end
	until true
	return result;
end


function UpdateXml(root, data, key)
	local path  = data['path'];
	local head  = 0;
	local tail  = string.find(path, '/');
    if tail == nil then 	tail = string.len(path)
    else                	tail = tail - 1;    end
	local name  = string.sub(path, head, tail);
	local tpos  = string.find(name, '#');
	local ntype = (tpos ~= nil and string.sub(name, tpos + 1, tail)) or "";

	if not TrackOp(root:name() == name and root:attribute(key) == ntype, VerifyResult,
				    string.format('match root name[%s] or root type[%s]', name, tostring(ntype))
				   )
	then
		return false;
	end

    local node  = nil;
	while tail < string.len(path) do
		head = tail + 2;
		tail = string.find(path, '/', head);
		if tail == nil then tail = string.len(path)
        else                tail = tail - 1;    end

		name  = string.sub(path, head, tail);
		tpos  = string.find(name, '#');
		ntype = (tpos ~= nil and string.sub(name, tpos + 1, tail)) or "";
		name  = (tpos ~= nil and string.sub(name, 0, tpos - 1)) or name;

		local nodes = root:select(name);
		node = nil;
		for i, k in pairs(nodes) do
			if k:attribute(key) == ntype then
				node = k;
				break;
			end
		end
		if node == nil then node = root:append(name); end

		if not TrackOp(node, VerifyResult, string.format('find or append subnode[%s] with type[%s]', name, ntype)) then
			return false;
		end
		if ntype ~= "" then node:set_attr(key, ntype) end
		root = node;
	end

	if data['attributes'] ~= nil then
		for key, val in pairs(data['attributes']) do
			if not TrackOp(root:set_attr(key, val), VerifyResult,
							string.format("set attributes[%s] as value[%s]", key, val))
			then
				return false;
			end
		end
	end

	if data['text'] ~= nil and not TrackOp(root:set_text(data['text']), VerifyResult,
											string.format("set text[%s] of node", data['text']))
	then
		return false;
	end
	return true;
end


function ATM_VerifyVersion(productinfo, srcdir, dstdir)
	return LO;		-- no need version verification any more;
--[[
	local function _splitversion (str)
		local result = {};

		local head = 0;
		local tail = string.find(str, '%.', head);
		if tail == nil then return end

        tail = tail - 1;
		while tail <= string.len(str) do
			local val = tonumber(string.sub(str, head, tail));
			if val == nil then
				break;
			end
			result[#result + 1] = val;

			head = tail + 2;
			tail = string.find(str, '%.', head);
			if tail == nil then tail = string.len(str);
			else                tail = tail - 1;        end
		end

		return result;
	end

	local srcdoc = rapidxml.open (string.format("%s\\%s", srcdir, productinfo));
	local dstdoc = rapidxml.open (string.format("%s\\%s", dstdir, productinfo));

	if not TrackOp(srcdoc ~= nil and srcdoc:root() ~= nil, VerifyResult,
				    string.format("open productinfo [%s\\%s]", srcdir, productinfo))
	then
		donelastop = false;
		return FATAL;
	end

	if dstdoc == nil or dstdoc:root() == nil then
		return LO;
	end

	local srcver = _splitversion(srcdoc:root():attribute('version'));
	local dstver = _splitversion(srcdoc:root():attribute('version'));

	if not TrackOp(#srcver == 4, VerifyResult, string.format("verify format of product version")) then
		donelastop = false;
		return FATAL;
	end

    for i = 1, 4, 1 do
		if dstver[i] == nil or dstver[i] < srcver[i] then
			return LO;
		elseif dstver[i] > srcver[i] then
			return HI;
		end
	end
	return EQ;
]]--
end



-- ************************************************************
-- Transaction Operations
-- ************************************************************

function TSA_StopService (svcname, fullpath, isdriver)
	redolist:push({'TSA_StopService', svcname, fullpath});
	if not donelastop then
		return false;
	end
	if isdriver == nil then isdriver = false end

	if Service.IsExistedService(svcname) == true then
		if not Service.StopService(svcname, 60 * 1000, true) then
			if isdriver then
				Service.SetServiceFailureAction(svcname, 0, nil, 0, 0, 0, 0, 0, 0);							-- SC_ACTION_NONE
			end
			killed = TrackOp(Threading.KillProcess(fullpath), VerifyResult,
							 string.format("stop service out of time and kill the process[%s : %s]",svcname, fullpath)
							)
			if isdriver then
				Service.SetServiceFailureAction(svcname, 0xffffffff, nil, 0x01, 0, 0x01, 0, 0x01, 0);		-- SC_ACTION_RESTART
			end

			if not killed then
				donelastop = false;
				return false;
			end
		end
	end
	redolist:pop();

	-- no undo info
	return true;
end


function TSA_InstallService (svcname, description, destdir, isdriver, starttype)
	redolist:push({'TSA_InstallService', svcname});
	if not donelastop then
		return false;
	end
	if isdriver == nil then isdriver = false end

--[[
	if Service.IsExistedService(svcname) ~= true
	   and not TrackOp(Service.InstallService(svcname, description, 0x10, 0x02, destdir),
						VerifyResult, string.format("install service[%s] located [%s]", svcname, destdir)
					   )
	then
		donelastop = false;
		return false;
	end
]]--
	local servicetype 	= 0x10						-- SERVICE_WIN32_OWN_PROCESS
	if starttype == nil then starttype = 0x02 end 	-- SERVICE_AUTO_START
	if isdriver then
		servicetype 	= 0x01	-- SERVICE_KERNEL_DRIVER
		starttype		= 0x01	-- SERVICE_SYSTEM_START
	end

	if not TrackOp(Service.InstallService(svcname, description, servicetype, starttype, destdir),
					VerifyResult, string.format("install service[%s] located [%s]", svcname, destdir)
				   )
	then
		donelastop = false;
		return false;
	end

	if not isdriver then
		TrackOp(Service.SetServiceFailureAction(svcname, 0xffffffff, nil, 0x01, 0, 0x01, 0, 0x00, 0),		-- SC_ACTION_NONE
				VerifyResult, string.format("set service failure actions[%s]", svcname));
	end

	redolist:pop();
	undolist:push({TSA_UninstallService, string.format("\"%s\"", svcname)})
	return true;
end


function TSA_StartService (svcname, args)
	if notexistaserror == nil then
		notexistaserror = true
	end

    redolist:push({'TSA_StartService', svcname, args});
	if not donelastop then
		return false;
	end

	local count = 0;
	while count < 30 do
		if Service.StartService(svcname, args) then break end
		Threading.Sleep(1000);
		count = count + 1;
	end
	if not TrackOp(count < 30, VerifyResult, string.format("start service[%s][%s]", svcname, tostring(count))) then
		donelastop = false;
		return false;
	end
	Threading.Sleep(2000);

	redolist:pop();
	-- no undo info
	return true;
end


function TSA_UninstallService (svcname, fullpath)
	redolist:push({'TSA_UninstallService', svcname});
	if not donelastop then
		return false;
	end

	-- stop service, and whatever it is stoped, try to uninstall the service;
	if Service.IsExistedService(svcname) then
		if not TrackOp(Service.StopService(svcname, 60 * 1000, true), VerifyResult, string.format("stop service[%s]", svcname)) then
			TrackOp(Threading.KillProcess(fullpath), VerifyResult, string.format("stop service out of time and kill the process[%s : %s]",svcname, fullpath))
		end
		if not TrackOp(Service.UninstallService(svcname), VerifyResult,string.format("uninstall service[%s]", svcname)) then
			donelastop = false;
		else
			donelastop = true;
		end
	else
		donelastop = true;
	end
	redolist:pop()

	-- nothing need to undo
	return true;
end


function TSA_RegisteIISService ()
	redolist:push({'TSA_RegisteIISService'});
	if not donelastop then
		return false;
	end

	-- track but not change donelastop
	if not Service.IsExistedService("aspnet_state") then
		local qlink   	= winfunc.WinQuickLink();
		local winpath 	= tostring(qlink.GetWindowsDir());
		local framework	= "Framework"
		if ATM_LoadTempData('BITs') == '64-bit' then
			framework = "Framework64"
		end
		TrackOp(ATM_Execute(string.format("%s\\Microsoft.NET\\%s\\v2.0.50727\\aspnet_regiis.exe", winpath, framework), " -i", {0, 259}, true),
				VerifyResult, string.format("registe IIS Service")
			   );
	end
	TrackOp(Service.SetServiceAutoStart("aspnet_state"), VerifyResult, "set aspnet_state as AutoStart");
	TrackOp(Service.SetServiceAutoStart("W3SVC"), 		 VerifyResult, "set W3SVC as AutoStart");
	redolist:pop()

	-- nothing need to undo
	return true;
end


function TSA_UpdateComponentFiles(categoryfile, srcdir, destdir, update, watchfilelist)
	g_WatchedFileUpdate = {};
	if watchfilelist == nil then
		watchfilelist = {}
	end

	if not donelastop then
		return false;
	end

	-- I. load category
	local category 	= LoadCategory(string.format("%s\\%s", srcdir, categoryfile));
	if not TrackOp(#category > 0, VerifyResult, string.format("get file list from category file [%s\\%s]", srcdir, categoryfile)) then
	    donelastop = false;
		return false;
	end

	local oldcategory = {}
	if FileSys.PathFileExists(string.format("%s\\%s", destdir, categoryfile)) then
		oldcategory = LoadCategory(string.format("%s\\%s", destdir, categoryfile));
	end
--	ATM_CleanMoveFileExTracks(destdir);

	-- II. copy file, just redo copy if copy failed, and undo if checksum not matched
	local newfilelist = {}
    for index, target in pairs(category) do
		src = string.format('%s\\%s\\%s', srcdir,  target['path'], target['name']);
		dst = string.format('%s\\%s\\%s', destdir, target['path'], target['name']);

		-- У鰲װļMD5
		local src_md5 = MD5.CheckSumFile_MD5(src);
		if not (src_md5 == target['checksum']) then
			RecordErrOp ("FAILED TO", string.format("match checksum of file[%s]", target['name']))
			donelastop = false;
			return false;
		end
		--[[
		if not TrackOp(src_md5 == target['checksum'], VerifyResult, string.format("match checksum of file[%s]", target['name'])) then
		    donelastop = false;
			return false;
		end
		--]]

		-- ignore copy if source and destination are same
		local dst_md5 = MD5.CheckSumFile_MD5(dst);
		if not (dst_md5 == src_md5) then
			if not (update and target['datafile'] == "1") then
				if ATM_CopyFile(src, dst) then
					for i, k in pairs(watchfilelist) do
						if k == target['name'] then g_WatchedFileUpdate[k] = true end
					end
				end
            end
		end

		-- remove file in old file list
		newfilelist[target['name']] = true;
	end

	-- III. delete old file according by two categories
	for index, target in pairs(oldcategory) do
		if newfilelist[target['name']] == nil then
			ATM_DelPath(string.format('%s\\%s\\%s', destdir, target['path'], target['name']), false);
			for i, k in pairs(watchfilelist) do
				if k == target['name'] then g_WatchedFileUpdate[k] = true end
			end
		end
	end

	-- IV.  update category file
	ATM_CopyFile(string.format("%s\\%s", srcdir, categoryfile), string.format("%s\\%s", destdir, categoryfile));

	return donelastop;
end


function TSA_RemoveComponentFiles(dir)
	redolist:push({'TSA_RemoveComponentFiles', dir});
	if not donelastop then
		return false;
	end
	ATM_DelPath(dir);
	redolist:pop();

	-- no undoes
	return donelastop;
end


function TSA_UpdateConfiguration(data, src, dest, update, key, utf8)
	if update and FileSys.PathFileExists(dest) then
		return true;
	end

	if key == nil then key = 'type' end
	repeat
		redolist:push({'TSA_UpdateConfiguration', _tostring("", data), src, dest});
		if not donelastop then break; end

		local doc = rapidxml.open (src);
		if not TrackOp(doc ~= nil and doc:root() ~= nil, VerifyResult, string.format("open configuration[%s]", src)) then break end
		redolist:pop();

		for i, k in pairs(data) do
			if not UpdateXml(doc:root(), k, key) then
				donelastop = false;
				RecordErrOp("FAILED TO", string.format("update xml[%s] to xml[%s]", src, dest));
				return false;
			end
		end
		
		if utf8 then
			if not TrackOp(doc:saveutf8(dest, true), VerifyResult, string.format("updated data into file[%s]", dest)) then break; end
		else
			if not TrackOp(doc:save(dest, true), VerifyResult, string.format("updated data into file[%s]", dest)) then break; end
		end

		undolist:push(ATM_DelPath, string.format("\"%s\"", dest));
		return true;
	until true

	donelastop = false;
	return false;
end


function TSA_InstallComponetConfigToEP (compname, compdir, epdir)
	srcfile = string.format("%s\\Jccfg_%s.xml", compdir, compname)
	dstfile = string.format("%s\\J_ClientCfg\\Jccfg_%s.xml", epdir, compname)
	if FileSys.PathFileExists(srcfile) then
		ATM_CopyFile(srcfile, dstfile);
	else
		ATM_DelPath (dstfile);
	end
	for i, suffix in pairs(component_configfilesuffixies) do
		srcfile = string.format("%s\\%s.%s", compdir, compname, suffix)
		dstfile = string.format("%s\\%s.%s", epdir  , compname, suffix)
		if FileSys.PathFileExists(srcfile) then
			ATM_CopyFile(srcfile, dstfile);
		else
			ATM_DelPath (dstfile);
		end
	end
end


function TSA_UninstallComponetConfigUnderEP (compname, epdir)
	ATM_DelPath(string.format("%s\\J_ClientCfg\\Jccfg_%s.xml", epdir, compname));
	for i, suffix in pairs(component_configfilesuffixies) do
		ATM_DelPath(string.format("%s\\%s.%s", epdir, compname, suffix));
	end
end

function TSA_InstallComponetPolicytoEp (id, compdir, epdir)
	srcfile = string.format("%s\\%s", compdir, id)
	dstfile = string.format("%s\\ptfile\\%s", epdir, id)
	if FileSys.PathFileExists(srcfile) then
		ATM_CopyFile(srcfile, dstfile);
	else
		ATM_DelPath (dstfile);
	end
end

function TSA_UninstallComponetPolicyToEp ( id, epdir )
	ATM_DelPath(string.format("%s\\ptfile\\%s", epdir, id));
end

function TSA_InstallAutoRun (key, value, type)
    redolist:push({'TSA_InstallAutoRun', key, value, type});
	if not donelastop or not ATM_SetRegKey('HKEY_LOCAL_MACHINE', 'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run', key, value, type) then
		donelastop = false;
		return false;
	end
    redolist:pop();

	undolist:push({ATM_RemoveRegKey, '\'HKEY_LOCAL_MACHINE\'', '\'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run\'', string.format("\"%s\"", key)});
	return true;
end


function TSA_UninstallRunOnce (key)
	redolist:push({'TSA_UninstallRunOnce', key, value, type});
	if not donelastop or not ATM_RemoveRegKey('HKEY_LOCAL_MACHINE', 'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run', key) then
		donelastop = false;
		return false;
	end
    redolist:pop();

	-- no undoes
	return true;
end



function TSA_ProcessWebmonBHO(dstdir, install)
	redolist:push({'TSA_InstallWebmonBHO', dstdir, install});
	if not donelastop then
		return false;
	end

	local qlink   = winfunc.WinQuickLink();
	local sysdir  = tostring(qlink.GetSystemDir());
	if install then
		local org_md5 = MD5.CheckSumFile_MD5(string.format("%s\\WebMon.dll",dstdir));
		local sys_md5 = MD5.CheckSumFile_MD5(string.format("%s\\WebMon.dll",sysdir));
		if sys_md5 ~= org_md5 then
			ATM_CopyFile(string.format("%s\\WebMon.dll",dstdir), string.format("%s\\WebMon.dll",sysdir))
		end
	end

	local comdll = winfunc.WinComDll();
	if not TrackOp(comdll.InstallCom(string.format("%s\\WebMon.dll",sysdir), install),
					VerifyResult, string.format("%s webmon bho[%s]", (install and "install") or "uninstall", component['name'])
				   )
	then
		donelastop = false;
		return false;
	end
	redolist:pop();

	if install then		-- do nothing if uninstall
		undolist:push({TSA_ProcessWebmonBHO, dstdir, not uninstall});
	end
	return true;
end

--[[
function TSA_ProcessLSP (dstdir, install)
	redolist:push({'TSA_InstallLSP', dstdir, install});
	if not donelastop then return false; end

	local ignored = FALSE;
	if install then ignored = SUCCESS end

	donelastop = false;
	local dll = winfunc.WinDll();
	repeat
	    if not TrackOp(dll.Load(string.format("%s\\rslspinst.dll", dstdir)),VerifyResult,
						string.format("load dll[%s\\rslspinst.dll]", dstdir)
					   )
		then break; end

		local err = dll.Invoke1("IsInstalled", "RSLSPMON");
		if not TrackOp(err ~= ERROR, VerifyResult, string.format("call entry[IsInstalled] of rslspinst.dll")) then break; end

		if err ~= ignored then
			if install then
				local qlink  = winfunc.WinQuickLink();
				local sysdir = tostring(qlink.GetSystemDir());
				local org_md5 = MD5.CheckSumFile_MD5(string.format("%s\\rslspmon.dll",dstdir));
				local sys_md5 = MD5.CheckSumFile_MD5(string.format("%s\\rslspmon.dll",sysdir));
				if sys_md5 ~= org_md5 then
					ATM_CopyFile(string.format("%s\\rslspmon.dll",dstdir), string.format("%s\\rslspmon.dll",sysdir))
				end

				err = dll.Invoke2("InstallRslspMon", "RSLSPMON", string.format("%s\\rslspmon.dll", sysdir))
			else
				err = dll.Invoke1("UninstallRslspMon", "RSLSPMON");
			end

			if not TrackOp(err ~= ERROR, VerifyResult,
                           string.format("call entry[%s] of rslspinst.dll", (install and "InstallRslspMon") or "UninstallRslspMon"))
            then break; end
			if not TrackOp(err ~= FALSE, VerifyResult,
                           string.format("%s LSP", (install and "install") or "uninstall"))
            then break; end
		end
		redolist:pop();

		undolist:push({TSA_ProcessLSP, dstdir, not install})
		donelastop = true;
	until true
	dll.Unload();

	return donelastop;
end
--]]

function TSA_UpdateWebSiteFiles(packages, writables, user, domain)
	redolist:push({'TSA_UpdateWebSiteFiles', packages, writables, user, domain});
	if not donelastop then
		return false;
	end

	local sitedirs = {}
	repeat
		-- uncompress package files
		for zipfile, basedir in pairs(packages) do
			if not donelastop then break; end
			ATM_UnzipFile(zipfile, basedir, nil);
			table.insert(sitedirs, basedir);
		end

		for path, right in pairs(writables) do
			if not donelastop then break; end
			ATM_ModifyPathSecurity(path, right, user, domain);
		end

		if not donelastop then break; end

		redolist:pop();
		undolist:push({TSA_RemoveWebSiteFiles, sitedirs});
	until true;

	return donelastop;
end


function TSA_RemoveWebSiteFiles(sitedir, vdirs)
	redolist:push({'TSA_RemoveWebSiteFiles', sitedirs});
	if not donelastop then
		return false;
	end

	for index, subdir in pairs(vdirs) do
		ATM_DelPath(string.format("%s\\%s", sitedir, subdir));
	end
	redolist:pop();

	-- undo nothing;
	return donelastop;
end

-- ж: ֱӵĿ¼жع, ɹķֵΪ0
function TSA_UninstallWebSite(sitename, sitedir, port, vdirbase, vdirs, toolfile, sslarg)
	redolist:push({'TSA_UninstallWebSite', sitename, sitedir, port, vdirbase, vdirs, toolfile});
	if not donelastop then
		return false;
	end

	if sslarg == nil then sslarg = "" end

	-- remove virtual directory and ignore errors
	for name, relatepath in pairs(vdirs) do
		ATM_Execute(toolfile, string.format(" -isdeploy 3 -sitepath \"%s\" -virtualname \"%s\" -physicalpath \"%s\\%s\" -port %s %s",
											sitedir, name, vdirbase, relatepath, tostring(port), sslarg
										   ),
				    {0}, true
				   );
	end
	redolist:pop();

	-- undo nothing;
	return donelastop;
end

-- װ : 1.  жĿ¼, ֱӵĿ¼İװ, жɹڲɣɹķֵΪ0
function TSA_InstallWebSite(sitename, sitedir, port, vdirbase, vdirs, toolfile, sslarg)
	redolist:push({'TSA_InstallWebSite', sitename, sitedir, port, vdirs, toolfile, sslarg});
	if not donelastop then
		return false;
	end

	if sslarg == nil then sslarg = "" end

	-- install virtual directory
	for name, relatepath in pairs(vdirs) do
		if not donelastop then break end

		-- Ŀ¼ೢ
		local args = string.format(" -isdeploy 1 -sitename \"%s\" -sitepath \"%s\" -virtualname \"%s\" -physicalpath \"%s\\%s\" -port %s %s",
									sitename, sitedir, name, vdirbase, relatepath, tostring(port), sslarg
								   );
		if not ATM_Execute(toolfile, args, {0},	true) then
			lasterror[#lasterror] = nil;
			donelastop = true;
			ATM_Execute(toolfile, args, {}, true);		-- Ϊ server2008 °װվδ֪浼´, ʱŹд
		end
	end

	if donelastop then
		redolist:pop();
		undolist:push({TSA_UninstallWebSite, sitename, sitedir, port, vdirs, toolfile, sslarg});
	end
	return donelastop;
end


--[[  1.  жվ, ڵķֵΪ 0, , ֱ/ֹͣվ
function TSA_RunWebSite(port, sitedir, toolfile, isstart)
	redolist:push({'TSA_RunWebSite', sitename, isstart, toolfile});
	if not donelastop then
		return false;
	end

	if not ATM_Execute(toolfile, string.format(" -isdeploy 6 -port %s -sitepath \"%s\"", tostring(port), sitedir), {0}, true) then
		lasterror[#lasterror] = nil;
		donelastop = true;
	else
		if isstart then
			ATM_Execute(toolfile, string.format(" -isdeploy 5 -port %s", tostring(port)), {0}, true);
		else
			ATM_Execute(toolfile, string.format(" -isdeploy 4 -port %s", tostring(port)), {0}, true);
		end
	end

	if donelastop then
		redolist:pop();
		undolist:push({TSA_RunWebSite, port, sitedir, toolfile, not isstart});
	end
	return donelastop;
end
--]]


function TSA_DeployCertificate(dstdir, srcdir, accessurl, password)
	redolist:push({'TSA_DeployCertificate', dstdir, accessurl, password});
	if not donelastop then
		return nil;
	end

	-- ɲװ֤
	local cafile = string.format("%s\\mgrca.pfx", dstdir);
	if not FileSys.PathFileExists(cafile) then
		if not TrackOp(ATM_Execute(string.format("%s\\rsca.exe", srcdir), string.format(" \"%s\" \"%s\" \"%s\"", password, cafile, accessurl), {0}, true),
						VerifyResult, string.format("generate certificate")
					   )
		then
			donelastop = false;
		end
	end

	local result = nil;
	if donelastop then
		local cert = winfunc.WinCertification();
		if not (TrackOp(cert.LoadCertFromPFX(cafile, password), VerifyResult, string.format("load cerificate[%s]", cafile))
				 and TrackOp(cert.ImportCertToSys("LOCAL_MACHINE", 	"MY", 	true), VerifyResult, string.format("import cerificate to LM My Store[%s]", cafile))
				 and TrackOp(cert.ImportCertToSys("LOCAL_MACHINE", 	"ROOT", true), VerifyResult, string.format("import cerificate to LM Root Store[%s]", cafile))
				 and TrackOp(cert.ImportCertToSys("CURRENT_USER",	"MY", 	true), VerifyResult, string.format("import cerificate to SC My Store[%s]", cafile))
				)
		then
			donelastop = false;
		else
			result = cert.GetCertSha1Hash();
		end
		cert.CloseCert()
	end

	if donelastop then
		redolist:pop();
	end
	return result;
end


function TSA_UnDeployCertificate(sha1hash)
	redolist:push({'TSA_UnDeployCertificate', sha1hash});
	if not donelastop then
		return false;
	end

	local cert = winfunc.WinCertification();
	if not (TrackOp(cert.RemoveCertFromSys("LOCAL_MACHINE", "MY", sha1hash), VerifyResult, string.format("remove cerificate[%s]", tostring(sha1hash)))
			 and TrackOp(cert.RemoveCertFromSys("LOCAL_MACHINE", "ROOT", sha1hash), VerifyResult, string.format("remove cerificate[%s]", tostring(sha1hash)))
			 and TrackOp(cert.RemoveCertFromSys("CURRENT_USER", "MY", sha1hash), VerifyResult, string.format("remove cerificate[%s]", tostring(sha1hash)))
		    )
	then
		donelastop = false;
	end

	if donelastop then
		redolist:pop();
	end
	return donelastop;
end


function TSA_InstallMSDE (srcdir, instancename, dbs, datadir, targetdir, pwd, port)
	redolist:push({'TSA_InstallMSDE', instancename, db, datadir, targetdir, pwd});
	if not donelastop then
		return false;
	end

	if not Service.IsExistedService(string.format("MSSQL$%s", instancename)) then
		ATM_Execute(string.format("%s\\setup.exe", srcdir),
					string.format(" SECURITYMODE=SQL sapwd=\"%s\" INSTANCENAME=\"%s\" DATADIR=\"%s\" TARGETDIR=\"%s\" DISABLENETWORKPROTOCOLS=0", pwd, instancename, datadir, targetdir),
					{0}, true, EXECUTE_WAITALWAYS
				   );
		local regkeypath = string.format('SOFTWARE\\Microsoft\\Microsoft SQL Server\\%s\\MSSQLServer\\SuperSocketNetLib\\Tcp', instancename);
		ATM_SetRegKey('HKEY_LOCAL_MACHINE', regkeypath, "TcpDynamicPorts", "", 'REG_SZ');
		ATM_SetRegKey('HKEY_LOCAL_MACHINE', regkeypath, "TcpPort", tostring(port), 'REG_SZ');
	end

	TSA_StartService(string.format("MSSQL$%s", instancename));

	local windb = winfunc.WinADODB();
	local host  = winfunc.GetHostName();
	for index, dbname in pairs(dbs) do
		if not donelastop then break end
		if not TrackOp(windb.CreateDB(string.format("%s\\%s", host, instancename), "sa", pwd, dbname), VerifyResult,
						string.format("create database [%s]", dbname)
					   )
		then
			donelastop = false;
		end
	end

	if donelastop then
		redolist:pop();
		undolist:push({TSA_UninstallMSDE, instancename});
	end
	return donelastop;
end


function TSA_UninstallMSDE (instancename)
	redolist:push({'TSA_UninstallMSDE', instancename, datadir, targetdir, pwd});
	if not donelastop then
		return false;
	end

	repeat
		if not Service.IsExistedService(string.format("MSSQL$%s", instancename)) then
			break;
		end

		if not TrackOp(Service.StopService(string.format("MSSQL$%s", instancename), 60 * 1000, true),
						VerifyResult, string.format("stop service[MSSQL$%s]", instancename)
					   )
		then
			donelastop = false;
			break;
		end

		local hkey 	  	= "HKEY_LOCAL_MACHINE";
		local subkey	= string.format("SOFTWARE\\Microsoft\\Microsoft SQL Server\\%s\\Setup", instancename);
		if not TrackOp(RegKey.Open(hkey, subkey, "KEY_READ"), VerifyResult, string.format("open regkey[%s//%s]", hkey, subkey)) then
			donelastop = false;
			break;
		end
		local productcode = tostring(RegKey.QueryStringValue("ProductCode"));
		RegKey.Close();


		local dll = winfunc.WinDll();
		local ver = dll.GetMsiVersion();
		if not TrackOp(ver ~= -1, VerifyResult, string.format("get msi version")) then
			donelastop = false;
			break;
		end

		local qlink = winfunc.WinQuickLink();
		if ver <= 2 then
			ATM_Execute(string.format("%s\\msiexec.exe", tostring(qlink.GetSystemDir())), string.format(" /qb! /X%s", productcode), {0}, true);
		else
			ATM_Execute(string.format("%s\\msiexec.exe", tostring(qlink.GetSystemDir())), string.format(" /passive /norestart /X\"%s\"", productcode), {0}, true);
		end
	until true

	if donelastop then
		redolist:pop();
		-- no undo needed
	end
	return donelastop;
end

function TSA_RemoveAppFromFireWall (appfile)
	redolist:push({'TSA_RemoveAppFromFireWall', appfile});
	if not donelastop then
		return false;
	end

	local firewall = winfunc.WindowsFirewall()
	firewall.EnableApp(appfile, "", 0x03, 0x00)		-- NET_FW_PROFILE_TYPE_MAX, WINDOWSFW_REMOVE

	if donelastop then
		redolist:pop();
		-- no undo needed
	end
	return donelastop;
end

-- ************************************************************
-- Composite Functions Wrapper
-- ************************************************************

function addremove (srcdir, rootdir, acts)
	-- call update routine if category file exist
	-- call install routine otherwise
	if FileSys.PathFileExists(string.format("%s\\%s\\%s", rootdir, component['name'], component['category'])) then
		RecordOp("RECORD TO", "call addremove with mode[update]");
		acts[1](srcdir, rootdir);
	else
		RecordOp("RECORD TO", "call addremove with mode[install]");
		acts[0] (srcdir, rootdir);
	end
end


function canrunit (srcdir, rootdir, acts)
	local framework	= "not found"
	if RegKey.Open("HKEY_LOCAL_MACHINE", "SOFTWARE\\Microsoft\\NET Framework Setup\\NDP\\v3.5", "KEY_READ") then
		if RegKey.QueryDWORDValue("Install", 0) == 1 then framework = "found"; end
		RegKey.Close();
	end
	local iis		= Service.IsExistedService('W3SVC') and "found" or "not found"

	local osinfo 	= winfunc.GetSystemInfo();
	local _, _, osver, bits = string.find(osinfo, "([^,]+),(.+)")

	oscond    = true
	framecond = true
	iiscond	  = true
	RecordOp("SUCCEED TO", string.format("collected system info : OS = %s, Bits = %s, .NET = %s, IIS = %s]", osver, bits, framework, iis));
	if component['role'] == 'server' then
		oscond    = (osver ~= "Windows XP" and osver ~= "Vista" and osver ~= "Unsupported OS")
		framecond = (framework == "found")
		iiscond   = (iis == "found")
		if not TrackOp(oscond and framecond and iiscond, VerifyResult,
						"check system enviroment before installation[IIS, .Net FrameWork 3.5, Server OS]")
		then
			donelastop = false;
		end
	elseif component['role'] == 'monitor' then
		-- installation requirment updated at 2012.07.30
		--[[
		oscond    = (osver ~= "Server 2008" and osver ~= "Server 2008 R2" and osver ~= "Server 2003" and osver ~= "Unsupported OS" and bits ~= "64-bit")
		if not TrackOp(oscond and framecond and iiscond, VerifyResult,
						"check system enviroment before installation[32-bit non-Server OS]")
		then
			donelastop = false;
		end
		]]--
	elseif component['role'] == 'terminator' then
		-- installation requirment updated at 2012.07.30
		--[[
		oscond    = (osver ~= "Unsupported OS")
		if not TrackOp(oscond and framecond and iiscond, VerifyResult,
						"check system enviroment before installation[OS higher than XP/Server2003(include)]")
		then
			donelastop = false;
		end
		]]--
	else
		-- component['role'] == 'library'
		-- will be confirmed anytime
	end

	unsatisfied = ""
	if not oscond 		then unsatisfied = string.format("ϵͳ[汾δ֧]") 						 end
	if not framecond 	then unsatisfied = string.format("%s .Net FrameWork 3.5[δװ]", 	unsatisfied) end
	if not iiscond	  	then unsatisfied = string.format("%s IIS[δװ]", 				 	unsatisfied) end

	winfunc.ExchangeInfo(unsatisfied);
end


function VerifyComponentFiles(srcdir, categoryfile)
	local category 	= LoadCategory(string.format("%s\\%s", srcdir, categoryfile));
	if not TrackOp(#category > 0, VerifyResult, string.format("get file list from category file [%s\\%s]", srcdir, categoryfile)) then
		donelastop = false;
		return;
	end

	-- II. copy file, just redo copy if copy failed, and undo if checksum not matched
	donelastop = true
    for index, target in pairs(category) do
		--[[ 
		ʱ֤MD5ԵĻУļУ
		--]]
		if target['datafile'] ~= "1" then
			src = string.format('%s\\%s\\%s', srcdir,  target['path'], target['name']);

			-- fatal error without redo
			local src_md5 = MD5.CheckSumFile_MD5(src);
			DebugOutput(1, "basement", string.format("******* %s : %s : %s", target['name'], target['checksum'], tostring(src_md5)))

			if not TrackOp(src_md5 == target['checksum'], VerifyResult, string.format("match checksum of file[%s]", target['name'])) then
				donelastop = false;
			end
		end
	end
end


function VerifyServices(svclist)
	if not donelastop then
		return
	end

	for svcname, executable in pairs(svclist) do
		--SERVICE_AUTO_START, SERVICE_RUNNING
		if not TrackOp(Service.VerifyServiceInfomations(svcname, 0x02, 0x04), VerifyResult,
						string.format("Verify that service[%s] is AutoStart and Running", svcname)
					   )
		then
			donelastop = false
			break;
		end
	end
end


function collectparam_agent(srcdir, rootdir, acts)
	collectparam(srcdir, rootdir);
end

function runbefore_agent(srcdir, rootdir, acts)
	if runbefore then
		runbefore(srcdir, rootdir)
	end
end

function collectparam_skip_agent(dstdir, srcdir, acts)
	DebugOutput(0, 'basement', string.format("[%s]collectparam skiped if UPDATE or UNINSTALL", component['name']))
end


function collectparam_addremove_agent(dstdir, srcdir, acts)
	local xml = string.format("%s\\%s", dstdir, component['category'])
	if FileSys.PathFileExists(xml) then
		RecordOp("RECORD TO", string.format("call collectparam_skip with mode[update], file[%s] found", xml));
		collectparam_skip_agent(dstdir, srcdir);
	else
		RecordOp("RECORD TO", string.format("call collectparam with mode[install], file[%s] not found", xml));
		collectparam_agent(dstdir, srcdir);
	end
end


function uninstall_agent(srcdir, rootdir, acts)
	uninstall(srcdir, rootdir)
end


function verifysp_agent(srcdir, rootdir, acts)
	VerifyComponentFiles(string.format("%s\\%s", rootdir, component['name']), component['category'])
	--VerifyServices(component['svcname'] or {})

	if component['name'] == 'ep' then
		local svr 		= tonumber(ATM_LoadTempData("ServerComponent") or 0);
		local df_open	= ATM_LoadTempData("File_SelfProtect_Open");
		local osinfo 	= winfunc.GetSystemInfo();
		local _, _, osver, bits = string.find(osinfo, "([^,]+),(.+)")
		if bits ~= "64-bit" then
			if not Service.IsExistedService("hookesm") then
				donelastop = false
			else
				if (svr == 0 and (df_open == nil or df_open == "closed")) then
					donelastop = false
				elseif (svr ~= 0 and df_open == "opened") then
					donelastop = false
				end
			end
		else
			if Service.IsExistedService("hookesm") then
				donelastop = false
			end
		end
	end
end


function geterrorinfo_agent(srcdir, rootdir, acts)
	local info = lasterror[1] or "";
	winfunc.ExchangeInfo(info);
end


function DoAction (action, srcdir, rootdir, hwnd)
	local acts = {	[0]  	= install,
					[1]  	= update,
					[2]  	= install,		--[[ repair	   ]]--
					[3]  	= addremove,	--[[ addremove ]]--
					[4]  	= uninstall_agent,
					[0x0a] 	= collectparam_agent,
					[0x0b] 	= runbefore_agent,
					[0x0c] 	= run,
					[16] 	= 'install_afterreboot',
					[20] 	= 'uninstall_afertreboot',
					[0x1e] 	= collectparam_skip_agent,
					[32] 	= canrunit,
					[0x21] 	= verifysp_agent,
					[0x22] 	= geterrorinfo_agent,
					[0x2e] 	= collectparam_addremove_agent
				  };
	g_ParentHwnd  = hwnd;
	local entry = acts[action];

	if TrackOp(entry ~= nil, VerifyResult, string.format("invalid or unsupported action id [%d]", action)) then
		entry(srcdir, rootdir, acts);
	end

	if not donelastop then
--		if action ~= 32 then
--	    	TSA_ErrorReport(g_ParentHwnd, string.format("Error Report[%s]", component['name']), "Installation failed with following errors", lasterror);
--	    end

		DebugOutput(0, 'basement', string.format("Error Report[%s]:\n%s", component['name'], _tostring("\t", lasterror)))
		if doneabort then
			return ABORT
		end
		return FATAL;
	end

	if #lasterror > 0 then
--		TSA_ErrorReport(g_ParentHwnd, string.format("Warning Report[%s]", component['name']), "Installation succeeded with following warnings", lasterror);
		DebugOutput(0, 'basement', string.format("Warning Report[%s]:\n%s", component['name'], _tostring("\t", lasterror)))
	end

	if donerestart then
		if component['name'] ~= 'rba' then
--			TSA_ErrorReport(g_ParentHwnd, "Warning Report", "continue, but, requestion of restart from non-Rba component will be regarded as an error", {});
			DebugOutput(0, 'basement', "Warning Report : continue, but, requestion of restart from non-Rba component will be regarded as an error");
		end
		return WARNING;
	end

	return SUCCESS;
end


function DoAction2 (action, dstdir, srcdir, envdata, mode)
	local acts = {
					[0x01]  =  event_pre_handle_handler,		-- EVENT_PRE_HANDLE
					[0x02]	=  event_post_handle_handler,		-- EVENT_POST_HANDLE
					[0x03]	=  event_pre_param_handler,			-- EVENT_PRE_PARAM
				    [0x05]	=  event_pre_update_handler			-- EVENT_PRE_UPDATE
				  };

	if component['role'] ~= 'product' then
		DebugOutput(0, "basement", string.format("Warning Report : component[%s] is not a product", tostring(component['name'])))
		return WARNING;
	end

	Initialize();

	local entry = acts[action];
	if entry ~= nil then
		entry(dstdir, srcdir, envdata, mode);
	end

	if not donelastop then
		DebugOutput(0, 'basement', string.format("Error Report[%s]:\n%s", component['name'], _tostring("\t", lasterror)))
		return FATAL;
	end

	if #lasterror > 0 then
		DebugOutput(0, 'basement', string.format("Error Report[%s]:\n%s", component['name'], _tostring("\t", lasterror)))
	end
	return SUCCESS;
end


-- ************************************************************
-- Dialog Builders
-- ************************************************************

function TSA_OpenDBCollect (dbdata)
    ----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local localmachine_name 	= nil
	local default_value_cache	= nil
	local default_stats_cache	= nil
	local logdb_values_cache	= {}
	local logdb_status_cache	= {}
	local logdb_paths_cache		= {}
	local db_values_cache		= {}
	local db_status_cache		= {}
	local db_paths_cache		= {}
	local other_values_cache	= {}		--	servertype, logdbcheck

	local sql_code				= 0x10
	local msde_code				= 0x20
	local db_code				= 0x01
	local logdb_code			= 0x02
	local checked_code			= 0x100
	local unchecked_code		= 0x000
	local other_code			= 0x040
	local path_code				= 0x080

	windlg = winfunc.WinDialog();
	windlg.CreateDialog	("ݿϢ", 0, 0, 296, 299, 23);

	--- default values of internal data --
	function load_localmachine_name ()
		if localmachine_name == nil then
			local localips 	= winfunc.GetValidIPs();
			local curhwnd 	= windlg.WindowHandle();
			if TrackOp(type(localips) == 'string' and localips ~= "", VerifyResult, "fetch any valid host IPs(v4), set as hostname if failed") then
				localmachine_name = TSA_ComboIPCollect(curhwnd, "ֶIP, ΪMSDEѡһȷıIP", localips);
			else
				localmachine_name = winfunc.GetHostName();
			end
		end
		return localmachine_name;
	end

	-- statusset_type is combined by 3 hex numbers:
	--			the lowest  for db (1) 		or logdb(2)
	--			the second  for sql(1) 		or msde (2)
	--			the highest for checked(1)	or unchecked(0)
	function load_default_status (statusset_type)
		if default_stats_cache == nil then
			default_stats_cache = {}
			-- sql
			default_stats_cache [unchecked_code + sql_code + db_code   ]  = {DataSource = true,  	DefaultDB = true,  	User = true,   	PWD = true}
			default_stats_cache [checked_code 	+ sql_code + db_code   ]  = {DataSource = true,  	DefaultDB = true,  	User = true,   	PWD = true}
			default_stats_cache [unchecked_code + sql_code + logdb_code]  = {DataSource = true,  	DefaultDB = true,  	User = true,   	PWD = true}
			default_stats_cache [checked_code 	+ sql_code + logdb_code]  = {DataSource = false,	DefaultDB = false,	User = false,	PWD = false}
			-- msde
			default_stats_cache [unchecked_code + msde_code + db_code	] = {DataSource = false, 	DefaultDB = true,  	User = false,  	PWD = true}
			default_stats_cache [checked_code 	+ msde_code + db_code	] = {DataSource = false, 	DefaultDB = true,  	User = false,  	PWD = true}
			default_stats_cache [unchecked_code + msde_code + logdb_code] = {DataSource = false, 	DefaultDB = true, 	User = false, 	PWD = false}
			default_stats_cache [checked_code 	+ msde_code + logdb_code] = {DataSource = false,	DefaultDB = false,	User = false,	PWD = false}
		end
		return default_stats_cache[statusset_type];
	end


	-- valueset_type is combined by 2 bytes:
	--			the lowest  for db (1) 		or logdb(2)
	--			the highest for sql(1) 		or msde (2)
	function load_default_values (valueset_type)
		if default_value_cache == nil then
			default_value_cache = {}
			-- sql
			default_value_cache [unchecked_code + sql_code  + db_code   ] = {DataSource = "", DefaultDB = "esmdb", 		User = "", PWD = "", Driver = "{Sql Server}"}
			default_value_cache [checked_code 	+ sql_code  + db_code   ] = {DataSource = "", DefaultDB = "esmdb", 		User = "", PWD = "", Driver = "{Sql Server}"}
			default_value_cache [unchecked_code + sql_code  + logdb_code] = {DataSource = "", DefaultDB = "esmlogdb", 	User = "", PWD = "", Driver = "{Sql Server}"}
			default_value_cache [checked_code 	+ sql_code  + logdb_code] = {DataSource = "", DefaultDB = "", 			User = "", PWD = "", Driver = "{Sql Server}"}
			-- msde
			default_value_cache [checked_code 	+ msde_code + logdb_code] = {DataSource = "", DefaultDB = "", 			User = "", PWD = "", Driver = "{Sql Server}"}
			-- other
			default_value_cache [other_code] 			 				  = {servertype = 'SQL Server', logdbselected = 0x00, msde_port = 7000}
			-- path
			default_value_cache [path_code 					+ db_code	] = {DefaultDB  = 'GenericDB/DB#Center/DefaultDB',
																			 DataSource = 'GenericDB/DB#Center/DataSource',
																			 User	    = 'GenericDB/DB#Center/User',
																			 PWD		= 'GenericDB/DB#Center/PWD',
																			 Driver		= 'GenericDB/DB#Center/Driver'
																			}
			default_value_cache [path_code					+ logdb_code] = {DefaultDB  = 'GenericDB/DB#Log/DefaultDB',
																			 DataSource = 'GenericDB/DB#Log/DataSource',
																			 User	    = 'GenericDB/DB#Log/User',
																			 PWD		= 'GenericDB/DB#Log/PWD',
																			 Driver		= 'GenericDB/DB#Log/Driver'
																			}
		end

		if (
				(
					(valueset_type == unchecked_code + msde_code + db_code)
					or (valueset_type == checked_code   + msde_code + db_code)
					or (valueset_type == unchecked_code + msde_code + logdb_code)
				)
				and
				(default_value_cache [valueset_type] == nil)
			)
		then
			local port = winfunc.GetFreeTcpPort(default_value_cache[other_code]['msde_port'])
			if Service.IsExistedService(string.format("MSSQL$%s", MSDE_INSTANCE_NAME)) then
				local regkeypath 	= string.format('SOFTWARE\\Microsoft\\Microsoft SQL Server\\%s\\MSSQLServer\\SuperSocketNetLib\\Tcp', MSDE_INSTANCE_NAME);
				local regport 		= ATM_GetRegKey('HKEY_LOCAL_MACHINE', regkeypath, "TcpPort")
				if TrackOp(regport ~= nil and regport ~= "", VerifyResult, string.format("load msde instacestance port info in regsitry")) then
					port = regport;
				end
			end
			local server	= string.format("%s\\%s,%s",load_localmachine_name(), MSDE_INSTANCE_NAME, port)
			default_value_cache [unchecked_code + msde_code + db_code	] = {DataSource = server, DefaultDB = "esmdb", 		User = "sa", PWD = "", Driver = "{Sql Server}"}
			default_value_cache [checked_code 	+ msde_code + db_code	] = {DataSource = server, DefaultDB = "esmdb", 		User = "sa", PWD = "", Driver = "{Sql Server}"}
			default_value_cache [unchecked_code + msde_code + logdb_code] = {DataSource = server, DefaultDB = "esmlogdb", 	User = "sa", PWD = "", Driver = "{Sql Server}"}
			default_value_cache [other_code]['msde_port'] = port
		end
		return default_value_cache[valueset_type];
	end

	--- internal data cache operation --
	function set_logdb_status (dataset)
		for key, value in pairs(dataset) do
			logdb_status_cache[key] = value;
		end
	end

	function set_logdb_values (dataset)
		for key, value in pairs(dataset) do
			logdb_values_cache[key] = value;
		end
	end

	function set_db_status (dataset)
		for key, value in pairs(dataset) do
			db_status_cache[key] = value;
		end
	end

	function set_db_values (dataset)
		for key, value in pairs(dataset) do
			db_values_cache[key] = value;
		end
	end

	function set_check_status (value)
		other_values_cache['logdbselected'] = value
	end

	function set_radio_status (value)
		other_values_cache['servertype'] = value
	end

	function load_db_values ()
		sync_update_db_values()
		return db_values_cache;
	end

	function reverse_check_status ()
		if other_values_cache['logdbselected'] == 0x01 then
			other_values_cache['logdbselected'] = 0x00
		else
			other_values_cache['logdbselected'] = 0x01
		end
	end

	function isselectmsde ()
		return other_values_cache['servertype'] == 'MSDE'
	end

	function ischecked ()
		return tonumber(other_values_cache['logdbselected']) == 0x01
	end

	function sync_pwd_value (value)
		db_values_cache   ['PWD'] = value
		logdb_values_cache['PWD'] = value
	end

	----------------------------------------
	-- init internal data
	----------------------------------------

	_tablecopy(other_values_cache, load_default_values(other_code))
	if dbdata['DB_Dialog'] ~= nil then
		for name, value in pairs(dbdata['DB_Dialog']) do
			other_values_cache[name] = dbdata['DB_Dialog'][name]
		end
	end

	local mode =	((isselectmsde() and msde_code   ) or sql_code      )
				  +	((ischecked   () and checked_code) or unchecked_code)
	_tablecopy(logdb_values_cache, load_default_values(mode + logdb_code))
	_tablecopy(logdb_status_cache, load_default_status(mode + logdb_code))
	_tablecopy(logdb_paths_cache,  load_default_values(path_code + logdb_code))
	_tablecopy(db_values_cache,    load_default_values(mode + db_code))
	_tablecopy(db_status_cache,    load_default_status(mode + db_code))
	_tablecopy(db_paths_cache,     load_default_values(path_code + db_code))

	if dbdata['CenterDB'] ~= nil then
		for name, value in pairs(dbdata['CenterDB']) do
			if dbdata['CenterDB'][name]['text'] ~= nil then
				db_values_cache[name] = dbdata['CenterDB'][name]['text']
			end
			if dbdata['CenterDB'][name]['path'] ~= nil then
				db_paths_cache [name] = dbdata['CenterDB'][name]['path']
			end
		end
	end
	if dbdata['LogDB'] ~= nil then
		for name, value in pairs(dbdata['LogDB']) do
			if dbdata['LogDB'][name]['text'] ~= nil then
				logdb_values_cache[name] = dbdata['LogDB'][name]['text']
			end
			if dbdata['LogDB'][name]['path'] ~= nil then
				logdb_paths_cache [name] = dbdata['LogDB'][name]['path']
			end
		end
	end

	-- init dialog frame
	local dbitems  = {};
	local logitems = {};
	for i, k in pairs({DataSource = 60, DefaultDB = 81, User = 102}) do
		dbitems[i] = windlg.InsertEdit 	(db_values_cache[i], 96, k,  138, 12);
	end
	dbitems['PWD'] = windlg.InsertPassWord 	(db_values_cache['PWD'], 96, 121, 138, 12);

	for i, k in pairs({DataSource = 187, DefaultDB = 208, User = 228}) do
		logitems[i] = windlg.InsertEdit (logdb_values_cache[i], 96, k,  138, 12);
	end
	logitems['PWD'] = windlg.InsertPassWord (logdb_values_cache['PWD'], 96, 248, 138, 12);

    local static_text = {" ", "ݿ", "û:", ":", ":", "ݿ:", "û:", ":"};
    local static_pos  = {62,        83,       104,       123,     188,       209,       231,       249};
	for i = 1, #static_text, 1 do
        windlg.InsertStatic (static_text[i], 9 , static_pos[i], 72, 9);
    end

	windlg.InsertGroup		("ݿ", 		7, 43,  282, 105)
    windlg.InsertGroup		("־ݿ", 	7, 169, 282, 105)
    windlg.InsertGroup		("ݿ",     	7, 7,   282, 34)

	local sql_radio  = windlg.InsertRadio 		("SQL Server",		63,  17,  70,  13)
	local msde_radio = windlg.InsertRadio 		("MSDE",			161, 17,  70,  13)

	local confirm    = windlg.InsertPushButton	("ύ", 				  122, 278, 67,  14)
	local logdbcheck = windlg.InsertCheck 		("־ݿͬ",7,   153, 126, 12)

	local collectedinfo	= {};
	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	--- internal gui operations ---
	function update_logdb_part ()
		local ref = nil
		for name, value in pairs(logitems) do
			ref = windlg.GetControlRef  (value);
			windlg.SetControlWindowText (ref, logdb_values_cache[name]);
			windlg.EnableControlWindow  (ref, logdb_status_cache[name]);
			windlg.ReleaseControlRef    (ref);
		end
	end

	function update_db_part ()
		local ref = nil
		for name, value in pairs(dbitems) do
			ref = windlg.GetControlRef  (value);
			windlg.SetControlWindowText (ref, db_values_cache[name]);
			windlg.EnableControlWindow  (ref, db_status_cache[name]);
			windlg.ReleaseControlRef    (ref);
		end
	end

	function update_check_part ()
		local ref = windlg.GetButtonRef(logdbcheck);
		windlg.SetButtonCheck(ref, tonumber(other_values_cache['logdbselected']));
		windlg.ReleaseButtonRef(ref);
	end

	function sync_update_logdb_pwd ()
		local ref = windlg.GetButtonRef(dbitems['PWD']);
		sync_pwd_value(tostring(windlg.GetControlWindowText (ref)));
		windlg.ReleaseButtonRef(ref);

		ref = windlg.GetButtonRef(logitems['PWD']);
		windlg.SetControlWindowText (ref, logdb_values_cache['PWD']);
		windlg.ReleaseButtonRef(ref);
	end

	function sync_update_db_values()
		local ref     = nil
		local dataset = {}
		for name, value in pairs(dbitems) do
			ref = windlg.GetControlRef  (value);
			dataset[name] = tostring(windlg.GetControlWindowText (ref));
			windlg.ReleaseControlRef    (ref);
		end

		for key, value in pairs(dataset) do
			db_values_cache[key] = value;
		end
	end

	function sync_update_logdb_values()
		local ref     = nil
		local dataset = {}
		for name, value in pairs(logitems) do
			ref = windlg.GetControlRef  (value);
			dataset[name] = tostring(windlg.GetControlWindowText (ref));
			windlg.ReleaseControlRef    (ref);
		end

		for key, value in pairs(dataset) do
			logdb_values_cache[key] = value;
		end
	end

	function update_radio_part ()
		local ref = nil;
		if other_values_cache['servertype'] == "SQL Server" then
			ref = windlg.GetButtonRef(sql_radio);
		else
			ref = windlg.GetButtonRef(msde_radio);
		end
		windlg.SetButtonCheck(ref, 0x01);   -- BST_CHECKED
		windlg.ReleaseButtonRef(ref);
	end

	function update_wholething ()
		update_db_part   ();
		update_logdb_part();
		update_check_part();
		update_radio_part();
	end

	--- window event handler ---
	function OpenDBCollect_OnOk ()
		local empty   = {};
		local ref     = windlg.GetButtonRef(logdbcheck);
		local checked = windlg.GetButtonCheck(ref);
		local curhwnd = windlg.WindowHandle();
		windlg.ReleaseButtonRef(ref);

		sync_update_db_values()
		if ischecked() then
			_tablecopy(logdb_values_cache, db_values_cache)
		else
			sync_update_logdb_values()
		end


		for key, value in pairs(db_values_cache) do
			if value == "" then empty[#empty + 1] = string.format('ݿ:%s',key) end
		end
		if not ischecked() then
			for key, value in pairs(logdb_values_cache) do
				if value == "" then empty[#empty + 1] = string.format('־ݿ:%s',key) end
			end
		end
		if #empty > 0 then
			TSA_ErrorReport(curhwnd, "ݴ", "Ϊ", empty);
			return 0;
		end

		local connect = {};
		ref = windlg.GetControlRef(confirm);
		windlg.EnableControlWindow(ref, false);
		windlg.SetControlWindowText(ref, "");

		-- only check db validity if server type is SQL Server
		local checked = 2;
		for name, cache in pairs({CenterDB = db_values_cache, LogDB = logdb_values_cache}) do
			local windb  = winfunc.WinADODB();

			-- check instance
			local checkinstance = windb.DetectInstance(cache['DataSource'], cache['User'], cache['PWD']);
			if (not isselectmsde() and not checkinstance) then
				TSA_ErrorReport(curhwnd, string.format("%s ݴ", name), "ݲϷ޷ SQL Server ʵ", {cache['DataSource']});
				break;
			elseif (isselectmsde() and checkinstance) then
				if not TSA_ContinueConfirm(	curhwnd,
											string.format("%s ݳͻ", name),
											string.format("MSDE ʵ [%s] Ѵڣ\nʹô˷ʵǸʵ", cache['DataSource']),
											"", ""
										   )
				then
					break;
				end
			end

			-- check db
			local checkdb = checkinstance;
			if checkdb then
				checkdb = windb.DetectDB(cache['DataSource'], cache['DefaultDB'], cache['User'], cache['PWD']);
			end

			if (not isselectmsde() and not checkdb) then
				TSA_ErrorReport(curhwnd, string.format("%s ݴ", name), "ݲϷ޷ SQL Server ݿ", {cache['DefaultDB']});
				break;
			elseif (isselectmsde() and checkdb) then
				if not TSA_ContinueConfirm(	curhwnd,
											string.format("%s ݳͻ", name),
										    string.format("MSDE ݿ [%s] Ѵڣ\nʹôݿ⻹Ǹݿ⣿", cache['DefaultDB']),
											"", ""
										   )
				then
					break;
				end
			end
			checked = checked - 1;
		end
		windlg.SetControlWindowText(ref, "ύ");
		windlg.EnableControlWindow(ref, true);
		windlg.ReleaseControlRef(ref);

		if checked > 0 then return 0; end

		for subname, setpair in pairs({CenterDB		= {vset = db_values_cache, 	  pset = db_paths_cache   },
										LogDB   	= {vset = logdb_values_cache, pset = logdb_paths_cache}
									   }
									  )
		do
			collectedinfo[subname] = {}
			for key, value in pairs(setpair['vset']) do
				collectedinfo[subname][key] = {}
				collectedinfo[subname][key]['text'] = value
				collectedinfo[subname][key]['path'] = setpair['pset'][key]
			end
		end
		collectedinfo['DB_Dialog'] = {}
		for key, value in pairs(other_values_cache) do
			collectedinfo['DB_Dialog'][key] = value
		end

		windlg.CloseDialog();
		return 1
	end

	function OpenDBCollect_OnMSDESelect()
		if not isselectmsde() then
			set_db_values   (load_default_values (unchecked_code + msde_code + db_code))
			set_db_status   (load_default_status (unchecked_code + msde_code + db_code))
			set_logdb_values(load_default_values (unchecked_code + msde_code + logdb_code))
			set_logdb_status(load_default_status (unchecked_code + msde_code + logdb_code))
			set_check_status(0x00);		-- BST_UNCHECKED
			set_radio_status('MSDE');

			update_wholething();
		end

		return 1;
	end

	function OpenDBCollect_OnLogdbCheck ()
		reverse_check_status()
		local mode =  	((isselectmsde() and msde_code   ) or sql_code      )
					  +	((ischecked   () and checked_code) or unchecked_code)
					  + logdb_code
		set_logdb_status(load_default_status(mode))
		if ischecked() then
			set_logdb_values(load_default_values(mode))
		else
			if isselectmsde() then
				local d1 = load_db_values()
				local d2 = load_default_values(mode)
				d1['DefaultDB'] = d2['DefaultDB']
				set_logdb_values(d1)
			else
				local d1 = load_default_values(mode)
				set_logdb_values(d1)
			end
		end
		update_logdb_part()
		return 1;
	end

	function OpenDBCollect_OnSQLSelect()
		if isselectmsde() then
			set_db_values   (load_default_values(unchecked_code + sql_code + db_code))
			set_db_status   (load_default_status(unchecked_code + sql_code + db_code))
			set_logdb_values(load_default_values(unchecked_code + sql_code + logdb_code))
			set_logdb_status(load_default_status(unchecked_code + sql_code + logdb_code))
			set_check_status(0x00);			-- BST_UNCHECKED
			set_radio_status('SQL Server');

			update_wholething();
		end
		return 1;
	end

	function OpenDBCollect_OnInit()
		windlg.RegisterCommandHandler(dbitems['PWD'],  0x300,  "OpenDBCollect_OnPWDChange") -- EN_CHANGE
		update_wholething()
		if ischecked() then
			OpenDBCollect_OnLogdbCheck()
		end
		return 1;
	end

	function OpenDBCollect_OnPWDChange ()
		if isselectmsde() and not ischecked() then
			sync_update_logdb_pwd();
		end
		return 1;
	end

	function OpenDBCollect_OnClose ()
		local curhwnd = windlg.WindowHandle();
		if not TSA_ContinueConfirm(curhwnd, "Ƿװ", "뱻رգװֹװ", "", "ֹ") then
			donelastop = false;		-- cancel following operations
			RecordErrOp("WARNING TO", "ûȡװ");
			windlg.CloseDialog();

			doneabort = true;
		end
		return 1;
	end
	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------

	windlg.RegisterCommandHandler  (logdbcheck, 	 0, 	 "OpenDBCollect_OnLogdbCheck") 	-- BN_CLICKED
	windlg.RegisterCommandHandler  (sql_radio,  	 0, 	 "OpenDBCollect_OnSQLSelect") 	-- BN_CLICKED
	windlg.RegisterCommandHandler  (msde_radio, 	 0, 	 "OpenDBCollect_OnMSDESelect") 	-- BN_CLICKED
	windlg.RegisterCommandIDHandler(confirm, 	"OpenDBCollect_OnOk")
	windlg.RegisterMessageHandler  (0x10, 	 	"OpenDBCollect_OnClose") 					--[[WM_CLOSE]]--
	windlg.RegisterMessageHandler  (0x0110, 	"OpenDBCollect_OnInit") 					--[[WM_INITDIALOG]]--

	windlg.Run(g_ParentHwnd, SW_SHOWNORMAL, global.LuaVM);

	return collectedinfo;
end


function TSA_OpenLinkCollect(caption, linkdata, data_tag, centername, centertype, centerport)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	-- !! check whether data is collected, which prelimited collected parameters (EP_LINK, BUS_LINK)
	local collected = ATM_LoadTempData("CollectedParams")
	if collected ~= nil and collected[data_tag] ~= nil  then
		return {IP = linkdata['text'], Port = linkdata['attributes']['Port']}
	end

	if linkdata == nil 									then linkdata = {} end
	if linkdata['text'] == nil							then linkdata['text'] = "" end
	if linkdata['attributes'] == nil 					then linkdata['attributes'] = {} end
	if tonumber(linkdata['attributes']['Port']) == nil 	then linkdata['attributes']['Port'] = "" end

	local windlg = winfunc.WinDialog();
	windlg.CreateDialog	(caption, 0, 0, 260, 77, 7);

    windlg.InsertStatic 	("˿:", 		184,24,	23,	11);
	windlg.InsertStatic 	("ַ:", 9,	24,	57,	11);

	local confirm 	= windlg.InsertPushButton	("ύ",		195,56, 50, 14);
	local detect	= windlg.InsertPushButton	(string.format("Զ%s", centername), 110, 56, 70, 14);
	local ip      	= windlg.InsertEdit 		(tostring(linkdata['text']), 				73,	22,	103,12);
	local port	  	= windlg.InsertEdit 		(tostring(linkdata['attributes']['Port']), 	215,22,	35,	12);
	local warning 	= windlg.InsertStatic    	("",            75, 42, 171,12);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	local collectedinfo = {};
	function OpenLinkCollect_OnOk ()
	    local err 	 = "";
	    repeat
	        local ref  = windlg.GetEditRef(ip);
	        local info = (ref ~= 0 and tostring(windlg.GetControlWindowText(ref))) or "";
	        windlg.ReleaseEditRef(ref);

			if not (IPChecker.CheckIPv4(info) or IPChecker.CheckIPv6(info)) then
				err = "IPַʽȷ";
	            break;
			end
	        collectedinfo['IP'] = info;

	        ref  = windlg.GetEditRef(port);
	        info = tonumber((ref ~= 0 and tostring(windlg.GetControlWindowText(ref))) or "");
	        windlg.ReleaseEditRef(ref);
	        if not (info ~= nil and info > 0 and info < 65535) then
	            err = "invalid format of port, please input again";
	            break;
	        end
	        collectedinfo['Port'] = info;

	        windlg.CloseDialog();
		    return 1;
	    until true;

        local ref2 = windlg.GetControlRef(warning);
        windlg.SetControlWindowText(ref2, err);
        windlg.ReleaseEditRef(ref2);
        return 1;
	end


	function OpenLinkCollect_OnClose ()
		local curhwnd = windlg.WindowHandle();
		if not TSA_ContinueConfirm(curhwnd, "Ƿװ", "뱻رգװֹװ", "", "ֹ") then
			donelastop = false;		-- cancel following operations
			RecordErrOp("WARNING TO", "ûȡװ");
			windlg.CloseDialog();

			doneabort = true;
		end
		return 1;
	end


	function OpenLinkCollect_OnDetect ()
		local centerdetector	= winfunc.WinCenterDetector();
		local curhwnd 		  	= windlg.WindowHandle();
		centerdetector.SetCerterInfo(centertype, 0, centerport, 1);

		TSA_StatusReport (curhwnd, "ڼЧϼ ", "δЧϼ", "Чϼ",
						  centerdetector.UDP_Detect_Threading_Start,
						  centerdetector.UDP_Detect_Threading_IsExited,
						  centerdetector.UDP_Detect_Threading_IsSucceed
						 );
		local list = centerdetector.GetFoundCenters();
		local addr = TSA_ComboIPCollect(curhwnd, "ֶЧϼģѡַ", list);
		if addr ~= "" then
			local _, _, ipstr, portstr = string.find(tostring(addr), "(%d+%.%d+%.%d+%.%d+):(%d+)");
			local ref = windlg.GetControlRef(ip);
			windlg.SetControlWindowText(ref, ipstr or "");
			windlg.ReleaseEditRef(ref);

			ref = windlg.GetControlRef(port);
			windlg.SetControlWindowText(ref, portstr or "");
			windlg.ReleaseEditRef(ref);
		end
		return 1;
	end

	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(confirm, "OpenLinkCollect_OnOk")
	windlg.RegisterCommandIDHandler(detect,  "OpenLinkCollect_OnDetect")
	windlg.RegisterMessageHandler  (0x10, 	 "OpenLinkCollect_OnClose")--[[WM_CLOSE]]--

	windlg.Run(g_ParentHwnd, SW_SHOWNORMAL, global.LuaVM);

	return collectedinfo;
end


function TSA_ServicePortCollect(caption, portdata, data_tag)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local collected = ATM_LoadTempData("CollectedParams")
	if collected ~= nil and collected[data_tag] ~= nil  then
		return {path = 'UTM/RsBaseComm/Common/Local/tcp', text = portdata['text']};
	end

	if portdata == nil			then portdata = {} end
	if portdata['text'] == nil	then portdata['text'] = "" end


	local windlg = winfunc.WinDialog();
	windlg.CreateDialog	(caption, 0, 0, 233, 59, 4);
    windlg.InsertStatic ("ΪָĶ˿", 20, 17, 71, 11);

	local confirm 	= windlg.InsertPushButton	("ύ",						176, 38, 50, 14);
	local port	  	= windlg.InsertNumberControl(tostring(portdata['text']), 	129, 15, 58, 13);
	local warning 	= windlg.InsertStatic    	("", 				 			15,  39,147, 13);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	local collectedinfo = {};
	function ServicePortCollect_OnOk ()
	    local err = "";
	    repeat
	        local ref  = windlg.GetEditRef(port);
	        info = tonumber((ref ~= 0 and tostring(windlg.GetControlWindowText(ref))) or "");
	        windlg.ReleaseEditRef(ref);
	        if not (info ~= nil and info > 0 and info < 65535) then
	            err = "invalid format of port, please input again";
	            break;
	        end
	        collectedinfo['text'] = info;
			collectedinfo['path'] = 'UTM/RsBaseComm/Common/Local/tcp';

	        windlg.CloseDialog();

		    return 1;
	    until true;

        local ref2 = windlg.GetControlRef(warning);
        windlg.SetControlWindowText(ref2, err);
        windlg.ReleaseEditRef(ref2);
        return 1;
	end


	function ServicePortCollect_OnClose ()
		local curhwnd = windlg.WindowHandle();
		if not TSA_ContinueConfirm(curhwnd, "Ƿװ", "뱻رգװֹװ", "", "ֹ") then
			donelastop = false;		-- cancel following operations
			RecordErrOp("WARNING TO", "ûȡװ");
			windlg.CloseDialog();

			doneabort = true;
		end
		return 1;
	end
	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(confirm, "ServicePortCollect_OnOk")
	windlg.RegisterMessageHandler  (0x10, 	 "ServicePortCollect_OnClose")	--WM_CLOSE--

	windlg.Run(g_ParentHwnd, SW_SHOWNORMAL, global.LuaVM);

	return collectedinfo;
end



function TSA_WebSiteInfo (websitename, data, onlyhttp, data_tag)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local collected = ATM_LoadTempData("CollectedParams")
	if collected ~= nil and collected[data_tag] ~= nil then
		return data;
	end

	local firstlocalip = SelectFirstLocalIP();

	if data 				== nil	then data 		= {}     end
	if data['porthttp']    	== nil	then data['porthttp']    	= 80  end
	if data['porthttps']   	== nil 	then data['porthttps']   	= 443 end
	if data['domain'] 		== nil 	then data['domain']		 	= firstlocalip end
	if data['websitetype'] 	== nil 	then data['websitetype'] 	= 'usedefault' end

	local windlg     = winfunc.WinDialog();
	windlg.CreateDialog	(string.format(" %s վϢ", websitename), 0, 0, 295, 151, 14);

    local static_text = {"HTTP˿", "HTTPS˿", "HTTPS", "ʵַΪ"};
    local static_pos  = {{16, 78, 41, 9}, {157, 79, 39, 9}, {16, 97, 55, 9}, {15, 114, 55, 9}};
	for i = 1, #static_text, 1 do
        windlg.InsertStatic (static_text[i], static_pos[i][1] , static_pos[i][2], static_pos[i][3], static_pos[i][4]);
    end
	windlg.InsertGroup ("վģʽ", 9,	11,271, 50);
	windlg.InsertGroup ("վ", 	9,  66,271, 61);

	local porthttp 		= windlg.InsertNumberControl (tostring(data['porthttp']),    77, 76, 41,  12);
	local porthttps 	= windlg.InsertNumberControl (tostring(data['porthttps']),   209,76, 43,  12);
	local domain	 	= windlg.InsertEdit   		 (tostring(data['domain']), 77, 95, 112, 12);

	local usedefault	= windlg.InsertRadio ("ʹĬվ  ֧HTTP",  15, 23, 129, 10);
	local httponly		= windlg.InsertRadio ("ʹԶվ㣨֧HTTP",  157,23, 129, 10);
	local httpsonly		= windlg.InsertRadio ("ʹԶվ㣨֧HTTPS", 15, 42, 151, 10);

	local urlnotify		= windlg.InsertStatic("", 77, 113, 193, 9);

	local confirm 		= windlg.InsertPushButton	("ύ",  					235, 130, 50, 14);

    local collected 	= {};
	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	function setinputstatus (switches)
		for i, k in pairs({porthttps, domain, porthttp}) do
			local ref = windlg.GetControlRef(k);
			windlg.EnableControlWindow (ref, switches[i]);
			windlg.ReleaseButtonRef(ref);
		end
	end

	function updateurl (protocol, domain, port, site)
		local ref = windlg.GetControlRef(urlnotify);
		windlg.SetControlWindowText (ref, string.format("%s://%s:%s/%s", protocol, domain, port, site));
		windlg.ReleaseControlRef(ref);
	end
	-----------------------------------------------
	local inited = false
	function WebSiteInfo_OnInit ()
		local ref;

		if onlyhttp then
			ref = windlg.GetButtonRef(httpsonly);
			windlg.EnableControlWindow(ref, false);
			windlg.ReleaseButtonRef(ref);
			if data['websitetype'] == 'httpsonly' then data['websitetype'] = 'usedefault' end
		end

		if 		data['websitetype'] == 'httpsonly' 	then
			ref = windlg.GetButtonRef(httpsonly);
			WebSiteInfo_OnOnlyHttps ();
		elseif	data['websitetype'] == 'httponly'	then
			ref = windlg.GetButtonRef(httponly);
			WebSiteInfo_OnOnlyHttp  ();
		else
			ref = windlg.GetButtonRef(usedefault);
			WebSiteInfo_OnUseDefault();
		end
		windlg.SetButtonCheck(ref, 0x01);   -- BST_CHECKED
		windlg.ReleaseButtonRef(ref);

		inited = true;
		return 1;
	end

	function WebSiteInfo_OnOnlyHttp ()
		collected["websitetype"] = 'httponly'
		data	 ["websitetype"] = 'httponly'
		setinputstatus ({false, false, true});

		local ref  		= windlg.GetControlRef(porthttp);
		local port 		= tostring(windlg.GetControlWindowText(ref));
		windlg.ReleaseControlRef(ref);

		updateurl	   ("http", firstlocalip, port, websitename);
		return 1;
	end

	function WebSiteInfo_OnOnlyHttps ()
		collected["websitetype"] = 'httpsonly'
		data	 ["websitetype"] = 'httpsonly'
		setinputstatus ({true, true, true});

		local ref  		= windlg.GetControlRef(porthttps);
		local port 		= tostring(windlg.GetControlWindowText(ref));
		windlg.ReleaseControlRef(ref);
		ref    			= windlg.GetControlRef(domain);
		local domain 	= tostring(windlg.GetControlWindowText(ref));
		windlg.ReleaseControlRef(ref);
		updateurl	   ("https", domain, port, websitename);
		return 1;
	end

	function WebSiteInfo_OnUseDefault ()
		collected["websitetype"] = 'usedefault'
		data     ["websitetype"] = 'usedefault'
		setinputstatus ({false, false, false});

		updateurl ("http", firstlocalip, "80", websitename);
		return 1;
	end

	function WebSiteInfo_OnClose ()
		local curhwnd = windlg.WindowHandle();
		if not TSA_ContinueConfirm(curhwnd, "Ƿװ", "뱻رգװֹװ", "", "ֹ")then
			donelastop = false;		-- cancel following operations
			RecordErrOp("WARNING TO", "ûȡװ");
			windlg.CloseDialog();

			doneabort = true;
		end
		return 1;
	end

	function WebSiteInfo_OnEditChange ()
		if not inited then return 1 end
		if 		data['websitetype'] == 'httpsonly' 	then 	WebSiteInfo_OnOnlyHttps ();
		elseif	data['websitetype'] == 'httponly'	then 	WebSiteInfo_OnOnlyHttp  ();
		else											  	WebSiteInfo_OnUseDefault(); end
		return 1;
	end

	function WebSiteInfo_OnOK ()
		local ref 		= nil;
		local curhwnd 	= windlg.WindowHandle();
		if data['websitetype'] == 'usedefault' then
			collected['porthttp'] = 80;
		else
			ref = windlg.GetEditRef(porthttp);
			collected['porthttp'] = tonumber(tostring(windlg.GetControlWindowText(ref))) or 0;
			windlg.ReleaseEditRef(ref);
			if collected['porthttp']  <= 0 or collected['porthttp']  > 65535 then
				TSA_ErrorReport(curhwnd, "ݴ", "ݷǷΪ", {"HTTP ˿"});
				return 0;
			end
		end

		if data['websitetype'] == 'httpsonly'	then
			local ref = windlg.GetEditRef(porthttps);
			collected['porthttps'] = tonumber(tostring(windlg.GetControlWindowText(ref))) or 0;
			windlg.ReleaseEditRef(ref);
			if collected['porthttps']  <= 0 or collected['porthttps']  > 65535 then
				TSA_ErrorReport(curhwnd, "ݴ", "ݷǷΪ", {"HTTPS ˿"});
				return 0;
			end

			ref = windlg.GetEditRef(domain);
			collected['domain'] = tostring(windlg.GetControlWindowText(ref));
			windlg.ReleaseEditRef(ref);
			if collected['domain'] == "" then
				TSA_ErrorReport(curhwnd, "ݴ", "ݷǷΪ", {"HTTPS "});
				return 0;
			end
		else
			collected['domain'] = firstlocalip;
		end

		windlg.CloseDialog();
		return 1;
	end
	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(confirm, "WebSiteInfo_OnOK")
	windlg.RegisterCommandHandler  (httponly,   0, "WebSiteInfo_OnOnlyHttp") 			-- BN_CLICKED
	windlg.RegisterCommandHandler  (httpsonly,  0, "WebSiteInfo_OnOnlyHttps") 			-- BN_CLICKED
	windlg.RegisterCommandHandler  (usedefault, 0, "WebSiteInfo_OnUseDefault") 			-- BN_CLICKED
	windlg.RegisterCommandHandler  (porthttp, 	0x300, 	"WebSiteInfo_OnEditChange") 	-- EN_CHANGE
	windlg.RegisterCommandHandler  (porthttps, 	0x300, 	"WebSiteInfo_OnEditChange") 	-- EN_CHANGE
	windlg.RegisterCommandHandler  (domain,		0x300,  "WebSiteInfo_OnEditChange") 	-- EN_CHANGE
	windlg.RegisterMessageHandler  (0x10, 	 "WebSiteInfo_OnClose")						--[[WM_CLOSE]]--
	windlg.RegisterMessageHandler  (0x0110,  "WebSiteInfo_OnInit") 						--[[WM_INITDIALOG]]--


	windlg.Run(g_ParentHwnd, SW_SHOWNORMAL, global.LuaVM);

	return collected;
end


function TSA_PathCollect(caption, linkdata, data_tag)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local collected = ATM_LoadTempData("CollectedParams")
	if collected ~= nil and collected[data_tag] ~= nil then
		return linkdata;
	end

	local windlg = winfunc.WinDialog();
	windlg.CreateDialog	(caption, 0, 0, 262, 58, 3);

	local pathedit 	= windlg.InsertEdit			(tostring(linkdata), 18, 15, 162, 13);
	local browse 	= windlg.InsertPushButton	("",	192, 14, 39, 14);
	local confirm	= windlg.InsertPushButton	("ύ", 	205, 37, 50, 14);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	local collectedinfo = nil;
	function PathCollect_OnInit ()
		local ref = windlg.GetControlRef(pathedit);
        windlg.SetControlWindowText(ref, tostring(linkdata));
        windlg.ReleaseControlRef(ref);
        return 1;
	end

	function PathCollect_OnBrowse ()
		local curhwnd 	= windlg.WindowHandle();
		local folder	= winfunc.BrowseFolderDialog(curhwnd);
		local ref 		= windlg.GetControlRef(pathedit);
		windlg.SetControlWindowText(ref, tostring(folder));
        windlg.ReleaseControlRef(ref);
		return 1;
	end

	function PathCollect_OnOk ()
		local ref 		= windlg.GetControlRef(pathedit);
		local curhwnd 	= windlg.WindowHandle();
        collectedinfo = tostring(windlg.GetControlWindowText(ref));
        windlg.ReleaseControlRef(ref);
		if collectedinfo == "" or not FileSys.PathFileExists(collectedinfo) then
			TSA_ErrorReport(curhwnd, "·", "·ǷΪ", {"Ŀ·"});
			return 0;
		end


		local freesize = winfunc.GetDiskFreeSpace(collectedinfo);
		local minsize  = 20 * 1024;
		if freesize < 20 * 1024 then
			if not TSA_ContinueConfirm(curhwnd, "ռ䲻",
										string.format("ռ䲻, С %d Gʹø·Ǹ·", minsize / 1024),
										"", ""
									   )
			then
				return 0;
			end
		end
		windlg.CloseDialog();
		return 1;
	end

	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(browse,  "PathCollect_OnBrowse")
	windlg.RegisterCommandIDHandler(confirm, "PathCollect_OnOk")
	windlg.RegisterMessageHandler  (0x0110,  "PathCollect_OnInit") --[[WM_INITDIALOG]]--

	windlg.Run(g_ParentHwnd, SW_SHOWNORMAL, global.LuaVM);

	return collectedinfo;
end


function TSA_ComboIPCollect(hwnd, caption, linkdata)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local _, p1v4, ip1v4 = string.find(tostring(linkdata), string.format("^%s,", ipv4_pattern));
	local _, _,    ip2v4 = string.find(tostring(linkdata), string.format("%s,",  ipv4_pattern), p1v4);
	local _, p1v6, ip1v6 = string.find(tostring(linkdata), string.format("^%s,", ipv6_pattern));
	local _, _,    ip2v6 = string.find(tostring(linkdata), string.format("%s,",  ipv6_pattern), p1v6);
	if (ip1v4 == nil or ip2v4 == nil) and (ip1v6 == nil or ip2v6 == nil) then
		RecordErrOp("FAILED TO", string.format("auto select one ip or can't found valid ip [%s]", tostring(ip1)));
		return (ip1v4 and tostring(ip1v4)) or (ip1v6 and tostring(ip1v6)) or "";
	end

	local windlg = winfunc.WinDialog();
	windlg.CreateDialog	(caption, 0, 0, 249, 57, 3);

    windlg.InsertStatic 	("ѡһЧIP",	17, 18, 64, 12);
	local confirm 	= windlg.InsertPushButton	("ύ",		192, 36, 50, 14);
	local ip      	= windlg.InsertComboBox		("1;2;3;4", 	91,  16, 119,92);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	local collectedinfo = nil;
	function ComboIPCollect_OnInit ()
		local ref = windlg.GetComboBoxRef(ip);
        windlg.SetComboBoxItems(ref, tostring(linkdata));
        windlg.ReleaseComboBoxRef(ref);
        return 1;
	end

	function ComboIPCollect_OnOk ()
		local ref = windlg.GetComboBoxRef(ip);
        collectedinfo = tostring(windlg.GetControlWindowText(ref));
        windlg.ReleaseComboBoxRef(ref);
		if collectedinfo ~= "" then
			windlg.CloseDialog();
			return 1;
		end
		return 0;
	end

	-- ************** deprived ****************
	--[[
	function ComboIPCollect_OnClose ()
		if not TSA_ContinueConfirm(curhwnd, "Ƿװ", "뱻رգװֹװ", "", "ֹ") then
			donelastop = false;		-- cancel following operations
			RecordErrOp("WARNING TO", "ûȡװ");
			windlg.CloseDialog();

			doneabort = true;
		end
		return 1;
	end
	]]
	-- ************** deprived ****************

	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(confirm, "ComboIPCollect_OnOk")
	windlg.RegisterMessageHandler  (0x0110,  "ComboIPCollect_OnInit") --[[WM_INITDIALOG]]--

	-- ************** deprived ****************
	-- windlg.RegisterMessageHandler  (0x10, 	 "ComboIPCollect_OnClose")--[[WM_CLOSE]]--
	-- ************** deprived ****************

	windlg.Run(hwnd, SW_SHOWNORMAL, global.LuaVM);

	return collectedinfo;
end


function TSA_ContinueConfirm (hwnd, caption, message, okstr, cancelstr)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local windlg = winfunc.WinDialog();
	windlg.CreateDialog	(caption, 0, 0, 262, 65, 3);

	local exit = windlg.InsertPushButton	(cancelstr, 205, 44, 50, 14);
	local goon = windlg.InsertPushButton	(okstr,  	147, 44, 50, 14);
	local warning = windlg.InsertStatic		(message,	18,  13, 225,26);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	local result = false;
	function ContinueConfirm_OnOk ()
		windlg.CloseDialog();
		result = true;
        return 1;
	end

	function ContinueConfirm_OnCancle ()
		windlg.CloseDialog();
		result = false;
        return 1;
	end

	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(goon, "ContinueConfirm_OnOk")
	windlg.RegisterCommandIDHandler(exit, "ContinueConfirm_OnCancle")

	windlg.Run(hwnd, SW_SHOWNORMAL, global.LuaVM);

	return result;
end


function TSA_ErrorReport (hwnd, caption, title, messages)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	local height = 0;
	local msg = string.format("%s:[%d]\n", title, #messages);
	for i, k in pairs(messages) do
		local head = 0;
		while head + 52 <= #k do
			msg  = string.format("%s\t%s\n", msg, string.sub(k, head, head + 52))
			head = head + 53;
			height = height + 1;
		end
		msg  = string.format("%s\t%s\n", msg, string.sub(k, head, #k))
		height = height + 1;
	end;
	height = height * 9;

	-- no alert if necessary
	-- if noalert then return end

	local windlg = winfunc.WinDialog();
	windlg.CreateDialog	(caption, 0, 0, 278, 43 + height, 2);

	local exit = windlg.InsertPushButton	("ȷ",  	205, 22 + height, 50, 14);
	local warning = windlg.InsertStatic    (""   ,	18,  13, 242, 8 + height);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	function ErrorReport_OnInit()
		local ref = windlg.GetControlRef(warning);
        windlg.SetControlWindowText(ref, msg);
        windlg.ReleaseControlRef(ref);
		return 1;
	end

	function ErrorReport_OnOk ()
		windlg.CloseDialog();
        return 1;
	end
	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterCommandIDHandler(exit, "ErrorReport_OnOk")
	windlg.RegisterMessageHandler  (0x0018, "ErrorReport_OnInit")--[[WM_SHOWWINDOW]]--

	windlg.Run(hwnd, SW_SHOWNORMAL, global.LuaVM);

	return true;
end


function TSA_StatusReport (hwnd, messages, failedmsg, succeedmsg, backproc_start, backproc_isexited, backproc_issucceed)
	----------------------------------------
	-- db-collection dialog building
	----------------------------------------
	function donothing ()
		return true;
	end

	if backproc_start    	== nil then backproc_start 	= donothing end
	if backproc_isexited	== nil then backproc_isexited	= donothing end
	if backproc_issucceed	== nil then backproc_issucceed	= donothing end

	local counter 	= 0;
	local steps		= "  ";
	local windlg   	= winfunc.WinDialog();
	windlg.CreateWhiteBoard	(0, 0, 198, 28, 1);
	local notify   	= windlg.InsertStatic (messages.."  ("..tostring(counter)..")", 27, 10, 153, 11);

	-------------------------------------
	-- begin of handler of db-collection dialog
	-------------------------------------
	function StatusReport_OnInit()
		windlg.StartTimer(1000);
		backproc_start();
		return 1;
	end

	function StatusReport_OnTimer ()
		local ref = windlg.GetControlRef(notify);
		counter = counter + 1;
		steps	= steps..". ";
		if counter % 5 == 0 then
			steps = "  ";
		end
        windlg.SetControlWindowText(ref, messages.."  ("..tostring(counter)..")"..steps);
        windlg.ReleaseControlRef(ref);

		if backproc_isexited() then
			windlg.SendMessage(0x10)		-- WM_CLOSE
		end
		return 1;
	end

	function StatusReport_OnClose ()
		local ref = windlg.GetControlRef(notify);
		if backproc_issucceed () then
			windlg.SetControlWindowText(ref, succeedmsg);
		else
			windlg.SetControlWindowText(ref, failedmsg);
		end
		windlg.ReleaseControlRef(ref);
		Threading.Sleep(1000);

		windlg.CloseDialog();
        return 1;
	end
	-------------------------------------
	-- end of handler of db-collection dialog
	-------------------------------------
	windlg.RegisterMessageHandler  (0x0110, "StatusReport_OnInit") 	--[[WM_INITDIALOG]]--
	windlg.RegisterMessageHandler  (0x0113, "StatusReport_OnTimer")	--[[WM_TIMER]]--
	windlg.RegisterMessageHandler  (0x10, 	"StatusReport_OnClose")	--[[WM_TIMER]]--

	windlg.Run(hwnd, SW_SHOWNORMAL, global.LuaVM);

	return true;
end


Initialize();
